from typing import NamedTuple, Optional, Union, Dict, List
from datetime import datetime, timedelta, date
import json

import redis
from bson import ObjectId

from robertcommonbasic.basic.dt.utils import DATETIME_FMT_FULL, DATETIME_FMT_DATE
from robertcommonbasic.basic.error.utils import InputDataError
from robertcommonbasic.basic.log import utils as logutils
from robertcommonbasic.basic.validation import input


class RedisConfig(NamedTuple):
    HOST: str
    PORT: int
    PWD: str
    LOG_ON_ERROR: bool = False
    STRONG_TYPE: bool = False


RedisJson = Union[int, float, bool, str, List, Dict, datetime, date, ObjectId]


class JsonRedis:
    class _SimpleJsonEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, datetime):
                return obj.strftime(DATETIME_FMT_FULL)
            elif isinstance(obj, date):
                return obj.strftime(DATETIME_FMT_DATE)
            else:
                return json.JSONEncoder.default(self, obj)

    class _StrongTypeHelper(json.JSONEncoder):
        _PREFIX_DATETIME = "::DATETIME::"
        _PREFIX_DATE = "::DATE::"
        _PREFIX_OBJID = "::OBJID::"

        def default(self, obj):
            text_value = self._try_convert_non_json_obj(obj)
            if text_value is None:
                return json.JSONEncoder.default(self, obj)
            else:
                return text_value

        @classmethod
        def _try_convert_non_json_obj(cls, obj):
            if isinstance(obj, datetime):
                return f"{cls._PREFIX_DATETIME}" \
                       f"{obj.strftime(DATETIME_FMT_FULL)}"
            elif isinstance(obj, date):
                return f"{cls._PREFIX_DATE}{obj.strftime(DATETIME_FMT_DATE)}"
            elif isinstance(obj, ObjectId):
                return f"{cls._PREFIX_OBJID}{obj}"
            else:
                return None

        @classmethod
        def _try_convert_non_json_str(cls, s: str):
            try:
                if s is None or not isinstance(s, str):
                    return None
                if s.startswith(cls._PREFIX_DATETIME):
                    return datetime.strptime(
                        s[len(cls._PREFIX_DATETIME):], DATETIME_FMT_FULL)
                if s.startswith(cls._PREFIX_DATE):
                    dt = datetime.strptime(
                        s[len(cls._PREFIX_DATE):], DATETIME_FMT_DATE)
                    return date(dt.year, dt.month, dt.day)
                if s.startswith(cls._PREFIX_OBJID):
                    return ObjectId(s[len(cls._PREFIX_OBJID):])
                return None
            except Exception:
                logutils.log_unhandled_error()
                raise

        @classmethod
        def _obj_hook(cls, j: dict):
            assert isinstance(j, dict)
            for key, value in j.items():
                if isinstance(value, str):
                    converted = cls._try_convert_non_json_str(value)
                    j[key] = value if converted is None else converted
                elif isinstance(value, list):
                    converted_list = []
                    for item in value:
                        converted = cls._try_convert_non_json_str(item)
                        converted_list.append(
                            item if converted is None else converted)
                    j[key] = converted_list
            return j

        @classmethod
        def convert_obj_to_str(cls, obj):
            converted = cls._try_convert_non_json_obj(obj)
            if converted is None:
                converted = json.dumps(obj, ensure_ascii=False, cls=cls)
            return converted

        @classmethod
        def convert_str_to_obj(cls, s):
            converted = cls._try_convert_non_json_str(s)
            if converted is None:
                try:
                    converted = json.loads(s, object_hook=cls._obj_hook)
                except Exception:
                    logutils.log_unhandled_error()
                    raise
            return converted

    StrongTypeHelper = _StrongTypeHelper

    def __init__(self, config: RedisConfig):
        self.host = input.ensure_not_none_str('host', config.HOST)
        self.port = input.ensure_not_none_int('port', config.PORT)
        self.pwd = input.ensure_not_none_str('pwd', config.PWD)
        self._log_on_error = bool(config.LOG_ON_ERROR)
        # if _strong_type is False (default), JsonRedis tries to convert
        # various types of values to string,
        # but not vice versa. To "get" method return strong typed values,
        # set _strong_type to True.
        self._strong_type = bool(config.STRONG_TYPE)

        try:
            self._redis = redis.StrictRedis(
                host=self.host, port=self.port, password=self.pwd)
        except Exception:
            if self._log_on_error:
                logutils.log_unhandled_error()
            raise

    @property
    def conn(self):
        return self._redis

    @property
    def strong_type(self):
        return self._strong_type

    def decode_value(self, redis_value: Optional[bytes]) -> Optional[RedisJson]:
        if not redis_value:
            value = None
        elif self._strong_type:
            value = self._StrongTypeHelper.convert_str_to_obj(
                redis_value.decode())
        else:
            value = json.loads(redis_value.decode())
        return value

    def encode_value(self, value: Optional[RedisJson]) -> Optional[bytes]:
        if self._strong_type:
            redis_value = self._StrongTypeHelper.convert_obj_to_str(value)
        else:
            if value is None:
                redis_value = None
            else:
                redis_value = json.dumps(
                    value, ensure_ascii=False, cls=self._SimpleJsonEncoder)
        return redis_value

    def get(self, name: str) -> Optional[RedisJson]:
        """
        Gets a REDIS object and try to load it into JSON object.
        :param name:
        :return:
        """
        name = input.ensure_not_none_str('name', name)
        try:
            rt = self._redis.get(name)
            if rt is None:
                return None
            value = self.decode_value(rt)
            return value
        except Exception:
            if self._log_on_error:
                logutils.log_unhandled_error()
            raise

    def expire(self, name: str, expire_seconds: Union[timedelta, int]) -> bool:
        """
        Expires a redis key in specified seconds.
        :param name:
        :param expire_seconds:
        :return: True if operation succeeds. False if the key does not exist.
        """
        key = input.ensure_not_none_str('name', name)
        if not isinstance(expire_seconds, timedelta):
            expire_seconds = input.ensure_not_none_int(
                'expire_seconds', expire_seconds)
        try:
            rt: bool = self._redis.expire(key, expire_seconds)
            return rt
        except Exception:
            if self._log_on_error:
                logutils.log_unhandled_error()
            raise

    def set(self, name: str, value: RedisJson,
            expire_period: timedelta = None) -> bool:
        """
        Set a redis with a JSON object.
        Note that the object will saved in text in REDIS.
        :param name:
        :param value:
        :param expire_period:
        :return: True if operation succeeds.
                 False if the operation condition is not met
                 (e.g. due to XX, NX option).
        """
        name = input.ensure_not_none_str('name', name)
        expire_period = input.ensure_of(
            'expire_period', expire_period, timedelta)
        ex = int(expire_period.total_seconds()) if expire_period else None
        try:
            value_to_set = self.encode_value(value)
            if value_to_set is None:
                # Treat set_to_none as delete operation
                self._redis.delete(name)
                return True
            rt: bool = self._redis.set(name, value_to_set, ex=ex)
            return rt
        except Exception:
            if self._log_on_error:
                logutils.log_unhandled_error()
            raise

    def delete(self, *names: str) -> int:
        """
        Try to delete multiple keys.
        :param names:
        :return: The number of keys successfully deleted.
        """
        keys = input.ensure_list_of(
            'names', list(names), input.ensure_not_none_str)
        try:
            rt: int = self._redis.delete(*keys)
            return rt
        except Exception:
            if self._log_on_error:
                logutils.log_unhandled_error()
            raise

    def keys(self, key_pattern: str) -> List[str]:
        """
        Try to get all keys matching key pattern.
        :return:
        """
        key_pattern = input.ensure_not_none_str('key_pattern', key_pattern)
        try:
            keys_bytes = self._redis.keys(key_pattern)
            rv = [key_byte.decode() for key_byte in keys_bytes]
            return rv
        except Exception:
            if self._log_on_error:
                logutils.log_unhandled_error()
            raise

    def hset(self, name: str, key: str, value: RedisJson) -> bool:
        name = input.ensure_not_none_str('name', name)
        key = input.ensure_not_none_str('key', key)

        try:
            redis_value = self.encode_value(value)
            rt: bool = bool(self._redis.hset(name, key, redis_value))
            return rt
        except Exception:
            if self._log_on_error:
                logutils.log_unhandled_error()
            raise

    def hmset(self, name: str, hash_map: Dict[str, RedisJson]) -> bool:
        name = input.ensure_not_none_str('name', name)
        if not isinstance(hash_map, dict):
            raise InputDataError(f"hash_map is not a dict, but {hash_map}")
        try:
            encoded_hash_map = {hash_key: self.encode_value(hash_value)
                                for hash_key, hash_value in hash_map.items()}
            rt: bool = bool(self._redis.hmset(name, encoded_hash_map))
            return rt
        except Exception:
            if self._log_on_error:
                logutils.log_unhandled_error()
            raise

    def hgetall(self, name: str) -> Dict[str, RedisJson]:
        name = input.ensure_not_none_str('name', name)
        try:
            rt = self._redis.hgetall(name)
            # An empty dict is returned if `name` does not exist
            assert rt is not None
            value = {k.decode(): self.decode_value(v) for k, v in rt.items()}
            return value
        except Exception:
            if self._log_on_error:
                logutils.log_unhandled_error()
            raise

    def hmget(self, name: str, keys: List[str]) -> List[Optional[RedisJson]]:
        name = input.ensure_not_none_str('name', name)
        keys = input.ensure_not_none_list_of(
            'keys', keys, input.ensure_not_none_str)
        try:
            rt = self._redis.hmget(name, keys)
            # An empty list is returned if `name` does not exist
            assert rt is not None
            values = [self.decode_value(v) for v in rt]
            return values
        except Exception:
            if self._log_on_error:
                logutils.log_unhandled_error()
            raise



import time
from typing import Optional, Any, Union, List, Dict

from sqlalchemy import create_engine, text
from sqlalchemy.pool.impl import QueuePool
from sqlalchemy.schema import CreateTable
from sqlalchemy.ext.declarative import declarative_base

import cx_Oracle
from datetime import datetime

import os
os.environ['NLS_LANG'] = 'SIMPLIFIED CHINESE_CHINA.UTF8'

class SQLAlCheAccessor:

    def __init__(self, config: Any=None):
        self.engine: dict = {}

    def add_engine(self, engine_name: str, engine_conn: str, engine_pool_class: Any=QueuePool, engine_pool_recycle: int=60):
        '''
                添加一个数据库连接池
                @Example:
                    PostgreSQL
                        # default
                        engine = create_engine('postgresql://scott:tiger@localhost/mydatabase')
                        # psycopg2
                        engine = create_engine('postgresql+psycopg2://scott:tiger@localhost/mydatabase')
                        # pg8000
                        engine = create_engine('postgresql+pg8000://scott:tiger@localhost/mydatabase')

                    MySQL
                        # default
                        engine = create_engine('mysql://scott:tiger@localhost/foo')
                        # mysqlclient (a maintained fork of MySQL-Python)
                        engine = create_engine('mysql+mysqldb://scott:tiger@localhost/foo')
                        # PyMySQL
                        engine = create_engine('mysql+pymysql://scott:tiger@localhost/foo')

                    Oracle
                        engine = create_engine('oracle://scott:tiger@127.0.0.1:1521/sidname')
                        engine = create_engine('oracle+cx_oracle://scott:tiger@tnsname')

                    SQL Server
                        # pyodbc
                        engine = create_engine('mssql+pyodbc://scott:tiger@mydsn')
                        # pymssql
                        engine = create_engine('mssql+pymssql://scott:tiger@hostname:port/dbname')

                    SQLite
                        engine = create_engine('sqlite:///foo.db')
                        engine = create_engine('sqlite:///C:\\path\\to\\foo.db')

            '''
        if engine_name not in self.engine.keys():
            engine = create_engine(engine_conn, poolclass=engine_pool_class, pool_recycle=engine_pool_recycle)
            if engine:
                self.engine[engine_name] = engine
        return self.engine.get(engine_name)

    def get_engine(self, engine_name: str):
        return self.engine.get(engine_name)

    def to_dict(self, records: Union[List[Optional[Dict]], Dict]):
        if isinstance(records, Dict):
            values = {}
            for k, v in records.items():
                values[k] = v
            return values
        elif isinstance(records, List):
            values = []
            for record in records:
                value = {}
                for k, v in record.items():
                    value[k] = v
                values.append(value)
            return values
        return records

    def read_sql(self, engine_name: str, sql_cmd: str, to_dict: bool=True):
        with self.get_engine(engine_name).connect() as conn:
            records = conn.execute(sql_cmd).mappings().all()
            if to_dict is True:
                return self.to_dict(records)
            return records


    def execute_sql(self, engine_name: str, sql_cmd: str, params: list=None):
        with self.get_engine(engine_name).connect() as conn:
            return conn.execute(text(sql_cmd), params).rowcount

    def execute_multi_sql(self, engine_name: str, cmd_tuple: list):
        with self.get_engine(engine_name).begin() as conn:
            rowcount = 0
            for cmd in cmd_tuple:
                rowcount = conn.execute(text(cmd[0]), cmd[1]).rowcount + rowcount
            return rowcount

    def get_table_struct(self, engine_name: str, table_name: str):
        engine = self.get_engine(engine_name)
        if engine is not None:
            base = declarative_base()
            base.metadata.reflect(engine)
            # 获取原表对象
            oldTable = base.metadata.tables[table_name]
            # 获取原表建表语句
            crate_sql = str(CreateTable(oldTable))
            base.metadata.clear()
            return crate_sql

    def generate_sql_format(self, table_name: str, action: str = 'replace', colums: list = None, filter: list = None):
        if action in ['replace', 'append']:
            return f"insert into {table_name} ({', '.join('{}'.format(k) for k in colums)}) VALUES ({', '.join(':{}'.format(k) for k in colums)})"
        elif action == 'update':
            return f"update {table_name} set {', '.join('{} = :{}'.format(k, k) for k in colums)} where {', '.join('{} = :{}'.format(k, k) for k in filter)}"
        elif action == 'delete':
            return f"delete from {table_name} where {', '.join('{} = :{}'.format(k, k) for k in filter)}"

    def generate_sql_cmds(self, table_name: str, records: list, action: str = 'replace', colums: list = None, filter: list = None):
        cmds = []
        if len(records) > 0:
            if action == 'replace':
                if filter is None:
                    cmds.append((f"delete from {table_name}", None))
                else:
                    cmds.append((self.generate_sql_format(table_name, 'delete', None, filter), records))

            cmds.append((self.generate_sql_format(table_name, action, records[0].keys() if colums is None else colums, filter), records))
        return cmds

class SynOracle():

    def __init__(self):
        self.accessor = None
        self.oracle_250 = 'oracle_250'
        self.oracle_216 = 'oracle_216'
        self.oracle_250_con = 'oracle://cy2003:goen_cy2003@10.192.1.250:1521/gc'
        self.oracle_216_con = 'oracle://cy2003:goen_cy2003@10.192.1.216:1521/gc'

        self.syn_interval = 600  # 同步间隔
        self.table_src = 'ACTION'  # 在场箱表
        self.table_dst = 'ACTION_SYN'
        self.cache_file = 'action_216_time.txt'
        self.limit = 500
        self.ole_last_syn = self.get_syn_time()  # 进出场扫描时间
        self.column = 'ID,CNTR_NO,CNTR_SIZE,CNTR_TYPE,MARK_OW,MARK_NS,OPERATOR_I,KIND_CNTR_I,PLAN_ID_I,KIND_I,STATUS_I,FE_I,EIR_NO_I,SEAL_NO_I,VESSEL_I,VOY_I,BLNO_I,CARGONAME_I,CLASS_I,UNDG_NO_I,GWET_I,PLACE_I,CUSER_I,DWDM_CD_I,TRUCK_NO_I,DRIVER_I,MARK_TC_I,REMARK_I,IC_KIND_I,IC_NO_I,DATE_IN,DATE_GATE_OUT,DATE_DROP,DRIVER_CY_I,SB_NO_I,CASH_I,LDR_I,DATE_LR_I,OPERATOR_O,KIND_CNTR_O,PLAN_ID_O,KIND_O,STATUS_O,FE_O,EIR_NO_O,SEAL_NO_O,VESSEL_O,VOY_O,BLNO_O,CARGONAME_O,CLASS_O,UNDG_NO_O,GWET_O,PLACE_O,CUSER_O,DWDM_CD_O,TRUCK_NO_O,DRIVER_O,MARK_TC_O,REMARK_O,IC_KIND_O,IC_NO_O,DATE_OUT,DATE_GATE_IN,DATE_LIFT,DRIVER_CY_O,SB_NO_O,CASH_O,LDR_O,DATE_LR_O,LOCK_OUT,SEAT_PLAN,SEAT,OPERATOR,KIND_CNTR,FE,STATUS,RELEASE_NO'.lower()
        self.columns = self.column.split(',')

        print(f"Start At {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}")

    def get_accessor(self):
        if self.accessor is None:
            self.accessor = SQLAlCheAccessor()
            self.accessor.add_engine(self.oracle_250, self.oracle_250_con)
            self.accessor.add_engine(self.oracle_216, self.oracle_216_con)
        return self.accessor

    def save_syn_time(self, now: datetime):
        self.ole_last_syn = now
        with open(self.cache_file, 'w') as f:
            f.write(f"{now.strftime('%Y-%m-%d %H:%M:%S')}")

    def get_syn_time(self):
        try:
            if os.path.exists(self.cache_file) is True:
                with open(self.cache_file, 'r') as f:
                    time = f.read()
                    if len(time) > 0:
                        return datetime.strptime(time, '%Y-%m-%d %H:%M:%S')
        except Exception as e:
            strError = 'ERROR: %s' % e.__str__()
            print(strError)
        return datetime.now()

    def split_to_list(self, values, size: int):
        if isinstance(values, dict):
            results = [{}]
            for k, v in values.items():
                if len(results[-1]) >= size:
                    results.append({k: v})
                else:
                    results[-1][k] = v
            return results
        elif isinstance(values, list):
            return [values[i:i + size] for i in range(0, len(values), size)]
        return values

    def syn_record(self):
        try:
            now = datetime.now()
            if self.get_accessor() is not None:
                records_on = self.accessor.read_sql(self.oracle_250, f"Select {self.column} from {self.table_src} where DATE_OUT is NULL")
                records_out = self.accessor.read_sql(self.oracle_250, f"Select {self.column} from {self.table_src} where DATE_OUT > to_date('{self.ole_last_syn.strftime('%Y-%m-%d %H:%M:%S')}', 'yyyy-mm-dd hh24:mi:ss') and DATE_OUT <= to_date('{now.strftime('%Y-%m-%d %H:%M:%S')}', 'yyyy-mm-dd hh24:mi:ss') order by DATE_OUT desc")
                print(f"[{now.strftime('%Y-%m-%d %H:%M:%S')}]: get online({len(records_on)}) out({len(records_out)}[{self.ole_last_syn.strftime('%Y-%m-%d %H:%M:%S')}-{now.strftime('%Y-%m-%d %H:%M:%S')}])")
                records_on.extend(records_out)
                if len(records_on) > 0:
                    records = self.split_to_list(records_on, self.limit)
                    cmds = []
                    for record in records:
                        cmds.extend(self.accessor.generate_sql_cmds(self.table_dst, record, 'replace', self.columns, ['id']))
                    result = self.accessor.execute_multi_sql(self.oracle_216, cmds)
                    if len(records_out) > 0:
                        self.save_syn_time(records_out[0]['date_out'])
                    end = datetime.now()
                    print(f"[{end.strftime('%Y-%m-%d %H:%M:%S')}]: syn success cost: {'{:.2f}'.format((end - now).total_seconds())}s online: {len(records_on)} out: {len(records_out)} result: {result}")
            else:
                raise Exception(f"no accessor")
        except Exception as e:
            print(e.__str__())

    def syn_records(self):
        while True:
            self.syn_record()
            time.sleep(self.syn_interval)

SynOracle().syn_records()


import asyncore
import logging
import sys
import socket
from abc import abstractmethod
from typing import NamedTuple, Any, Optional, Union
from enum import Enum
from threading import Thread, Lock, Event

from robertcommonbasic.basic.cls.utils import daemon_thread, function_thread
from robertcommonbasic.basic.data.frame import FRAMEFLAG, PACKAGETYPE, FRAMETYPE, TRANSFERTYPE, pack_frame, unpack_frame, get_pack_start_length, get_pack_header_length, convert_bytes_to_int
from robertcommonbasic.basic.data.conversion import format_bytes


class CloseException(Exception):
    pass


class NormalException(Exception):
    pass


class SocketType(Enum):
    TCP_CLIENT = 'tcp_client'
    TCP_SERVER = 'tcp_server'
    UDP_CLIENT = 'udp_client'
    UDP_SERVER = 'udp_server'


class SocketConfig(NamedTuple):
    MODE: SocketType
    HOST: str
    PORT: int = 9500
    POOL: int = 0
    BUFFER: int = 1400
    LISTEN: int = 500
    TIME_OUT: int = 10
    CALL_BACK: dict = {}
    HANDLE_CLASS: Any = None
    PARAENT_CLASS: Any = None


# 定制应答式处理类
class SocketHandler(asyncore.dispatcher_with_send):

    def __init__(self, config: SocketConfig, sock=None, addr=None):
        asyncore.dispatcher_with_send.__init__(sock)

        self.config = config
        self.sock = sock
        self.addr = addr
        self.valid = True

        self.start_length = get_pack_start_length()
        self.header_length = get_pack_header_length()

        if self.config.MODE == SocketType.TCP_CLIENT:
            self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
            self.connect((self.config.HOST, self.config.PORT))
        elif self.config.MODE == SocketType.UDP_CLIENT:
            self.create_socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.connect((self.config.HOST, self.config.PORT))

    def __str__(self):
        if isinstance(self.addr, tuple) and len(self.addr) >= 2:
            return f"{self.addr[0]}:{self.addr[1]}"
        return str(self.addr)

    def __del__(self):
        self.exit()

    def exit(self):
        self.valid = False
        self.close()

    def format_bytes(self, data: bytes) -> str:
        return ' '.join(["%02X" % x for x in data]).strip()

    def handle_read(self):
        try:
            data_start = self.recv(self.start_length)    # 包头
            if len(data_start) == self.start_length:
                if data_start == FRAMEFLAG.START_FLAG:    #
                    data_header = self.recv(self.header_length)    # 包头长度
                    if len(data_header) == self.header_length:
                        data_end_length = convert_bytes_to_int(data_header[-4:]) + self.start_length + 1
                        data_end = self.recv(data_end_length)  # 包内容
                        if len(data_end) == data_end_length:
                            if data_end[-2:] == FRAMEFLAG.END_FLAG:  # 包结尾
                                if 'analyze_data' in self.config.CALL_BACK.keys():
                                    self.config.CALL_BACK['analyze_data'](self, data_start + data_header + data_end)
        except Exception as e:
            logging.error(f"handle read({self}) fail ({e.__str__()})")

    def handle_error(self):
        t, e, trace = sys.exc_info()
        if 'handle_error' in self.config.CALL_BACK.keys():
            self.config.CALL_BACK['handle_error'](self, e)
        self.close()
        self.valid = False

    def handle_close(self):
        if 'handle_close' in self.config.CALL_BACK.keys():
            self.config.CALL_BACK['handle_close'](self)
        self.close()
        self.valid = False

    def handle_connect(self):    
        if 'handle_connect' in self.config.CALL_BACK.keys():
            self.config.CALL_BACK['handle_connect'](self)

    def writable(self):
        return True

    def handle_write(self):
        pass

    def readable(self):
        return True

    def check_invalid(self) -> bool:
        return self.valid


class SocketServer(asyncore.dispatcher):

    def __init__(self, config: SocketConfig):
        asyncore.dispatcher.__init__(self)
        self.config = config
        self.sockets = {}

        if self.config.MODE == SocketType.TCP_SERVER:
            self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        else:
            self.create_socket(socket.AF_INET, socket.SOCK_DGRAM)

        self.set_reuse_addr()
        self.bind((self.config.HOST, self.config.PORT))

        if self.config.MODE == SocketType.TCP_SERVER:
            self.listen(self.config.LISTEN)

    def handle_close(self):
        self.close()

    def handle_accept(self):
        client = self.accept()
        if client is not None:
            sock, addr = client
            if self.verify_request(sock, addr) is True:
                if self.config.HANDLE_CLASS is not None:
                    self.config.HANDLE_CLASS(self.config, sock, addr, self)
                else:
                    self.sockets[f"{addr[0]}:{addr[1]}"] = SocketHandler(self.config, sock, addr)

    def verify_request(self, sock, address):
        return True

    def clear_invalid(self):
        sock_addrs = list(self.sockets.keys())

        for sock_addr in sock_addrs:
            if sock_addr in self.sockets.keys() and self.sockets[sock_addr].check_invalid() is False:
                del self.sockets[sock_addr]


class SocketAccessor:

    def __init__(self, config: SocketConfig):
        self.config = config
        self.accessor = None
        self.thread = None

    def __del__(self):
        self.exit()

    def exit(self):
        if self.accessor:
            self.accessor.close()
        if self.thread:
            self.thread.join()
        self.accessor = None
        self.thread = None

    def start(self, thread: bool = True):
        if self.config.MODE in [SocketType.TCP_SERVER, SocketType.UDP_SERVER]:
            self.accessor = SocketServer(self.config)
        else:
            if self.config.HANDLE_CLASS is not None:
                self.accessor = self.config.HANDLE_CLASS(self.config)
            else:
                self.accessor = SocketHandler(self.config)

        if self.config.POOL == 0:
            if thread is True:
                self.thread = Thread(target=asyncore.loop)
            else:
                asyncore.loop()
        else:
            if thread is True:
                self.thread = Thread(target=asyncore.loop, kwargs={'timeout': self.config.TIME_OUT})
            else:
                asyncore.loop(self.config.TIME_OUT)

        if thread is True and self.thread:
            self.thread.start()


#####
class AsyncResult:

    def __init__(self, client, request_id: int, data: bytes, timeout: float, retries: int = 1):
        self.client = client
        self.request_id = request_id
        self.data = data
        self.retries = retries
        self.timeout = timeout
        self.wait_event = Event()
        self.error = None
        self.result = None

    def __del__(self):
        self.close()

    def close(self):
        self.wait_event.clear()

    def send_request(self, wait_config: bool = True) -> dict:
        for r in range(self.retries):
            self.send()
            if wait_config is False:
                if self.error:
                    raise self.error
                return {}
            else:
                if self.wait(self.timeout) is True:
                    return self.result
        if self.error:
            raise self.error
        raise Exception(f"no config")

    def recv(self, package: dict):
        if package.get('package_index') == self.request_id:
            self.wait_event.set()
        self.result = package

    def format_bytes(self, data: bytes) -> str:
        return ' '.join(["%02X" % x for x in data]).strip()

    def send(self):
        self.error = None
        try:
            if self.client:
                if self.client.send(self.data) <= 0:
                    raise Exception(f"send fail")
            else:
                raise Exception(f"no connect")
        except Exception as e:
            self.error = Exception(f"send exception({e.__str__()})")

    def wait(self, timeout: float) -> bool:
        while True:
            if not self.wait_event.wait(timeout):
                return False
            return True


class AsyncResultManage:
    
    def __init__(self, start_package_id: int, end_package_id: int, call_backs: dict):
        self.start_package_id = start_package_id
        self.end_package_id = end_package_id
        self.call_backs = call_backs
        
        self.package_id = start_package_id
        self.requests = {}
        
    def set_debug_log(self, content: str):
        if 'set_debug_log' in self.call_backs.keys():
            self.call_backs['set_debug_log'](content)

    def get_pack_id(self) -> int:
        if self.package_id >= self.end_package_id:
            self.package_id = self.start_package_id
        self.package_id = self.package_id + 1
        return self.package_id

    def config_request(self, request_id: int, result: dict = None):
        if request_id in self.requests.keys():
            if result is not None:
                self.set_debug_log(f"recv {request_id} (FRAMETYPE.CONFIG)")
                self.requests[request_id].recv(result)
            else:
                del self.requests[request_id] 
    
    def create_config(self, client, package_type: int, package_index: int, transfer_type: int, data: Any):
        self.set_debug_log(f"send {package_index}({FRAMETYPE(FRAMETYPE.CONFIG).__str__()}--{PACKAGETYPE(package_type).__str__()}--{TRANSFERTYPE(transfer_type).__str__()})")
        answer = dict()
        answer['frame_type'] = FRAMETYPE.CONFIG
        answer['package_type'] = package_type
        answer['package_index'] = package_index
        answer['transfer_type'] = transfer_type
        answer['data'] = data
        return client.send(pack_frame(answer))

    def create_request(self, client, package_index: Optional[int], frame_type: int, package_type: int, transfer_type: int, data, timeout: float, retries: int = 1, wait_config: bool = True):
        try:
            if package_index is None:
                package_index = self.get_pack_id()
            self.set_debug_log(f"send {package_index}({FRAMETYPE(frame_type).__str__()}--{PACKAGETYPE(package_type).__str__()}--{TRANSFERTYPE(transfer_type).__str__()})")
            self.requests[package_index] = AsyncResult(client, package_index, pack_frame({'frame_type': frame_type, 'package_type': package_type, 'package_index': package_index, 'transfer_type': transfer_type, 'data': data}), timeout, retries)
            return self.requests[package_index].send_request(wait_config)
        except Exception as e:
            logging.error(f"create_request({client})({e.__str__()})")
        finally:
            self.config_request(package_index)


class SocketIPHandler:

    def __init__(self, config: SocketConfig, sock=None, addr=None):
        self.config = config
        self.sock = sock
        self.addr = addr
        self.reg_tag = ''
        self.valid = True
        self.lock = Lock()

        self.start_length = get_pack_start_length()
        self.header_length = get_pack_header_length()
        self.exit_flag = False
        self.rec_buffer = b''

        self.receive().start()
        
        # 触发连接事件
        self.handle_connect()

    def __str__(self):
        if isinstance(self.addr, tuple) and len(self.addr) >= 2:
            return f"{self.addr[0]}:{self.addr[1]}"
        return str(self.addr)

    def __del__(self):
        self.exit()

    def exit(self):
        self.valid = False
        self.close()

    def format_bytes(self, data: bytes) -> str:
        return ' '.join(["%02X" % x for x in data]).strip()

    def close(self):
        self.valid = False
        self.exit_flag = True
        if self.sock:
            self.sock.close()
        self.sock = None

    def handle_connect(self):
        if 'handle_connect' in self.config.CALL_BACK.keys():
            self.config.CALL_BACK['handle_connect'](self)

    def handle_close(self, reason: str = ''):
        self.close()
        if 'handle_close' in self.config.CALL_BACK.keys():
            self.config.CALL_BACK['handle_close'](self, reason)

    def handle_error(self, e: Exception):
        self.close()
        if 'handle_error' in self.config.CALL_BACK.keys():
            self.config.CALL_BACK['handle_error'](self, e)
        else:
            logging.error(f"handle_error({self})({e.__str__()})")
            
    def recv_bytes(self, length: int) -> Optional[bytes]:
        data = b''
        while self.check_invalid() is True and len(data) < length:
            rec_length = min(self.config.BUFFER, length - len(data))
            rec_data = self.sock.recv(rec_length)
            if rec_data is None or len(rec_data) == 0:
                raise CloseException(f"remote close")
            data += rec_data
        if len(data) != length:
            raise NormalException(f"recv length fail({len(data)}/{length})")
        return data

    def recv_default(self):
        if self.recv_bytes(1) == FRAMEFLAG.START_FLAG[:1]:
            if self.recv_bytes(1) == FRAMEFLAG.START_FLAG[1:]:
                data_header = self.recv_bytes(self.header_length)  # 包头长度为9
                if data_header is not None:
                    data_end_length = convert_bytes_to_int(data_header[-4:]) + self.start_length + 1
                    data_end = self.recv_bytes(data_end_length)  # 包内容
                    if data_end is not None:
                        if data_end[-2:] == FRAMEFLAG.END_FLAG:  # 包结尾
                            try:
                                package = unpack_frame(FRAMEFLAG.START_FLAG + data_header + data_end)
                                if 'handle_data' in self.config.CALL_BACK.keys():
                                    self.config.CALL_BACK['handle_data'](self, package)
                            except Exception as e:
                                logging.error(f"unpack frame fail({e.__str__()})")

    @daemon_thread
    def receive(self):
        while self.exit_flag is False:
            try:
                if 'handle_read' in self.config.CALL_BACK.keys():
                    data = self.sock.recv(self.config.BUFFER)
                    if data is None or len(data) == 0:
                        raise CloseException(f"remote close")
                    elif data is not None and len(data) > 0:
                        self.config.CALL_BACK['handle_read'](self, data)
                else:
                    self.recv_default()
            except socket.timeout as e:
                pass
            except NormalException as e:
                logging.error(e.__str__())
            except CloseException as e:
                self.handle_close(e.__str__())
            except Exception as e:
                self.handle_error(e)

    def send(self, data: bytes):
        with self.lock:
            if self.sock:
                length = len(data)
                buffer = self.config.BUFFER
                group = int(length/buffer)
                for i in range(group):
                    self.sock.send(data[i*buffer: (i+1)*buffer])
                if group*buffer < length:
                    self.sock.send(data[group*buffer:])
                return length
            raise Exception(f"no connect")
    
    def check_invalid(self) -> bool:
        return self.valid


class SocketIPAccssor:

    def __init__(self, config: SocketConfig):
        self.config = config
        self.accessor = None
        self.exit_flag = False
        self.thread = None
        self.sockets = {}
        self.lock = Lock()

    def __del__(self):
        self.exit()

    def exit(self):
        self.exit_flag = True
        for k, v in self.sockets.items():
            v.close()
        if self.accessor:
            self.accessor.close()
        self.accessor = None
        self.thread = None
        self.sockets = {}

    def create_client(self):
        with self.lock:
            if self.config.MODE == SocketType.TCP_CLIENT:
                self.accessor = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            elif self.config.MODE == SocketType.UDP_CLIENT:
                self.accessor = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

            self.accessor.settimeout(self.config.TIME_OUT)  # 设置连接超时
            self.accessor.connect((self.config.HOST, self.config.PORT))
            self.accessor.settimeout(None)

            self.sockets[f"{self.config.HOST}:{self.config.PORT}"] = SocketIPHandler(self.config, self.accessor)
        
    def create_server(self):
        with self.lock:
            if self.config.MODE == SocketType.TCP_SERVER:
                self.accessor = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            elif self.config.MODE == SocketType.UDP_SERVER:
                self.accessor = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

            self.accessor.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.accessor.bind((self.config.HOST, self.config.PORT))

            if self.config.MODE == SocketType.TCP_SERVER:
                self.accessor.listen(self.config.LISTEN)

        while self.exit_flag is False:
            try:
                sock, addr = self.accessor.accept()  # 接收连接
                if self.verify_request(sock, addr) is True:
                    if self.config.TIME_OUT > 0:
                        sock.settimeout(self.config.TIME_OUT)  # 设置连接超时
                    self.sockets[f"{addr[0]}:{addr[1]}"] = SocketIPHandler(self.config, sock, addr)
            except Exception as e:
                pass

    def verify_request(self, sock, address):
        return True

    def clear_invalid(self):
        sock_addrs = list(self.sockets.keys())
        for sock_addr in sock_addrs:
            if sock_addr in self.sockets.keys() and self.sockets[sock_addr].check_invalid() is False:
                del self.sockets[sock_addr]

    def start(self, thread: bool = True):
        if self.config.MODE in [SocketType.TCP_SERVER, SocketType.UDP_SERVER]:
            if thread is True:
                self.thread = Thread(target=self.create_server)
            else:
                self.create_server()
        else:
            if thread is True:
                self.thread = Thread(target=self.create_client)
            else:
                self.create_client()

        if thread is True and self.thread:
            self.thread.start()


# IOT 重构
class IOTNetMessage:

    def __init__(self):
        self.heads: Optional[bytearray] = None
        self.contents: Optional[bytearray] = None
        self.sends: Optional[bytearray] = None

    def __str__(self):
        return f"IOTNetMessage"

    def get_head_length(self) -> int:
        """
            协议头数据长度，也即是第一次接收的数据长度
        """
        return 0

    def get_content_length(self) -> int:
        """二次接收的数据长度"""
        return 0

    def get_frame_length(self) -> int:
        """帧大小"""
        return 1024

    def check_response(self) -> bool:
        """回复报文校验"""
        return True

    def check_head(self) -> bool:
        """回复报文校验"""
        return True
    

class IOTNetResult:

    is_success: bool = False    # 是否成功的标志
    msg: str = 'Unknown'    # 操作返回的错误消息
    code: int = 10000   # 错误码
    contents: list = [None] * 20    # 结果数组

    '''结果对象类，可以携带额外的数据信息'''
    def __init__(self, code: int = 0, msg: str = ""):
        self.code = code
        self.msg = msg
        self.is_success = False
        self.contents: list = [None] * 20  # 结果数组

    def __str__(self):
        return f"code: {self.code} msg: {self.msg}"

    def copy(self, result):
        if result is not None and isinstance(result, IOTNetResult):
            self.code = result.code
            self.msg = result.msg

    @staticmethod
    def create_fail(result=None):
        failed = IOTNetResult()
        if result is not None:
            failed.code = result.code
            failed.msg = result.msg
        return failed

    @staticmethod
    def create_success(contents: Optional[list] = None):
        success = IOTNetResult()
        success.is_success = True
        success.msg = 'Success'
        if contents is not None and not isinstance(contents, list):
            contents = [contents]
        if isinstance(contents, list):
            for i, content in enumerate(contents):
                success.contents[i] = content
        return success


class IOTNetworkBase:

    def __init__(self):
        self.iot_socket = None
        self.socket_error = False
        self.callbacks = {}

    def __str__(self):
        return f"IOTNetworkBase"

    def __del__(self):
        self.exit()

    @abstractmethod
    def extra_on_close(self, sock: socket):
        """连接上服务器后需要进行的初始化操作"""
        raise NotImplementedError()

    def logging(self, **kwargs):
        if 'call_logging' in kwargs.keys():
            self.callbacks['call_logging'] = kwargs.get('call_logging')

        if 'content' in kwargs.keys():
            call_logging = self.callbacks.get('call_logging')
            if call_logging:
                call_logging(**kwargs)

    def exit(self):
        if self.iot_socket is not None:
            self.iot_socket.close()
        self.iot_socket = None

    def close_socket(self, sock: socket):
        if sock is not None:
            sock.close()
            self.extra_on_close(sock)

    def receive(self, sock: socket, length: Optional[int] = None, size: Optional[int] = 1024):
        """接收固定长度的字节数组"""
        if length == 0:
            return IOTNetResult.create_success([bytearray(0)])

        data = bytearray()
        try:
            if isinstance(length, int):
                while len(data) < length:
                    pack_size = size if length - len(data) > size else length - len(data)
                    d = sock.recv(pack_size)
                    if not d:
                        raise CloseException(f"close")
                    elif len(d) < pack_size:
                        data.extend(d)
                        break
                    else:
                        data.extend(d)
            else:
                data.extend(sock.recv(size))
            return IOTNetResult.create_success([data])
        except socket.timeout as e:
            return IOTNetResult(msg=f"receive fail({e.__str__()})")
        except (CloseException, Exception) as e:
            self.close_socket(sock)
            return IOTNetResult(msg=f"receive fail({e.__str__()})")

    def send(self, sock: socket, data: Optional[bytes]):
        """发送消息给套接字，直到完成的时候返回"""
        try:
            if data is not None:
                self.logging(content=f"send {len(data)}: [{format_bytes(data)}]")
                sock.send(data)
            return IOTNetResult.create_success()
        except Exception as e:
            self.close_socket(sock)
            return IOTNetResult(msg=f"send fail({e.__str__()})")

    def connection(self, host: str, port: int, timeout: Optional[Union[int, float]] = None, sock_type: int = 1):
        sock = socket.socket(socket.AF_INET, sock_type)
        try:
            sock.settimeout(10 if timeout is None else timeout)
            sock.connect((host, port))
            if timeout is None:
                sock.settimeout(None)
            return IOTNetResult.create_success([sock])
        except Exception as e:
            return IOTNetResult(msg=f"connect fail({e.__str__()})")

    def listen(self, host: str, port: int, size: int = 500, sock_type: int = 1):
        sock = socket.socket(socket.AF_INET, sock_type)
        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.bind((host, port))
            sock.listen(size)
            return IOTNetResult.create_success([sock])
        except Exception as e:
            return IOTNetResult(msg=f"bind fail({e.__str__()})")

    def receive_msg(self, sock: socket, net_msg):
        """接收一条完整的数据，使用异步接收完成，包含了指令头信息"""
        if net_msg is None:
            return self.receive(sock)

        if isinstance(net_msg, IOTNetMessage):
            result = IOTNetResult()
            head = self.receive(sock, net_msg.get_head_length(), net_msg.get_frame_length())
            if head.is_success is False:
                result.copy(head)
                return result

            net_msg.heads = head.contents[0]
            if net_msg.check_head() is False:
                result.msg = f"Receive Invalid Head"
                return result

            content_length = net_msg.get_content_length()
            if content_length == 0:
                net_msg.contents = bytearray(0)
            else:
                content_result = self.receive(sock, content_length, net_msg.get_frame_length())
                if content_result.is_success is False:
                    result.copy(content_result)
                    return result
                net_msg.contents = content_result.contents[0]

            if net_msg.contents is None:
                net_msg.contents = bytearray(0)

            if net_msg.check_response() is False:
                result.msg = f"Receive Invald package"
                return result
            return IOTNetResult.create_success([self.cat_bytes(net_msg.heads, net_msg.contents)])

    def cat_bytes(self, data1: bytes, data2: bytes):
        if data1 is None and data2 is None:
            return None
        if data1 is None:
            return data2
        if data2 is None:
            return data1
        return data1 + data2

    def get_connected(self, sock: socket) -> bool:
        if sock is not None:
            if getattr(sock, '_closed') is False:
                return True
        return False


class IOTNetwork(IOTNetworkBase):
    
    def __init__(self, host: str, port: int, timeout: Optional[Union[int, float]] = None, size: int = 500):
        super().__init__()
        self.host = host
        self.port = port
        self.timeout = timeout
        self.size = size
        self.lock = Lock()
        self.is_persistent: bool = True  # 是否长连接模式

    @abstractmethod
    def get_net_msg(self):
        """获取一个新的消息对象的方法"""
        raise NotImplementedError()

    @abstractmethod
    def extra_on_connect(self, sock: socket):
        """连接上服务器后需要进行的初始化操作"""
        raise NotImplementedError()

    @abstractmethod
    def extra_on_disconnect(self, sock: socket):
        """在将要和服务器进行断开的情况下额外的操作"""
        raise NotImplementedError()

    @abstractmethod
    def extra_on_receive(self, sock: socket, datas: bytes):
        raise NotImplementedError()

    def pack_command_with_header(self, command: bytearray):
        """对当前的命令进行打包处理，通常是携带命令头内容，标记当前的命令的长度信息，需要进行重写，否则默认不打包"""
        return command

    def unpack_response(self, send: bytearray, response: bytearray):
        """根据对方返回的报文命令，对命令进行基本的拆包，例如各种Modbus协议拆包为统一的核心报文，还支持对报文的验证"""
        return IOTNetResult.create_success([response])

    # Client
    def create_and_initialication(self):
        """连接并初始化网络套接字"""
        result = self.connection(self.host, self.port, self.timeout)
        if result.is_success:
            # 初始化
            initi = self.extra_on_connect(result.contents[0])
            if initi.is_success is False:
                if result.contents[0] is not None:
                    self.close_socket(result.contents[0])
                result.is_success = initi.is_success
                result.copy(initi)
        return result

    def connect(self):
        result = IOTNetResult()
        self.exit()
        con_result = self.create_and_initialication()
        if con_result.is_success is False:
            self.socket_error = True
            con_result.contents[0] = None
            result.msg = con_result.msg
        else:
            self.iot_socket = con_result.contents[0]
            result.is_success = True
        return result

    def disconnect(self):
        result = IOTNetResult()
        with self.lock:
            result = self.extra_on_disconnect(self.iot_socket)
            self.exit()
            return result

    # Server
    def start_server(self):
        result = IOTNetResult()
        self.exit()
        server_result = self.listen(self.host, self.port, self.size)
        if server_result.is_success is False:
            server_result.contents[0] = None
            result.msg = server_result.msg
        else:
            self.iot_socket = server_result.contents[0]
            function_thread(self.accept, True, self.iot_socket).start()
            result.is_success = True
        return result

    def accept(self, s: socket):
        while self.iot_socket is not None:
            sock, addr = s.accept()  # 接收连接
            function_thread(self.receive_client, True, sock).start()

    def stop_server(self):
        result = IOTNetResult()
        with self.lock:
            result = self.extra_on_disconnect(self.iot_socket)
            self.exit()
            return result

    def get_socket(self):
        """获取本次操作的可用的网络套接字"""
        if self.is_persistent:
            # 长连接模式
            if self.socket_error or self.iot_socket is None or self.get_connected(self.iot_socket) is False:
                connect = self.connect()
                if connect.is_success is False:
                    self.socket_error = True
                    return IOTNetResult(msg=connect.msg)
                else:
                    self.socket_error = False
                    return IOTNetResult.create_success([self.iot_socket])
            else:
                return IOTNetResult.create_success([self.iot_socket])
        else:
            # 短连接模式
            return self.create_and_initialication()

    def read_from_socket(self, s: socket, send: Optional[bytearray], has_response: bool = True, pack_unpack: bool = True):
        """在其他指定的套接字上，使用报文来通讯，传入需要发送的消息，返回一条完整的数据指令"""
        send_value = self.pack_command_with_header(send) if pack_unpack else send

        net_msg = self.get_net_msg()
        if net_msg is not None:
            net_msg.sends = send_value

        send_result = self.send(s, send_value)
        if send_result.is_success is False:
            return IOTNetResult.create_fail(send_result)

        if has_response is False:
            return IOTNetResult.create_success([bytearray(0)])

        # 接收数据信息
        result_receive = self.receive_msg(s, net_msg)
        if result_receive.is_success is False:
            return IOTNetResult(code=result_receive.code, msg=result_receive.msg)

        # 拼接结果数据
        return self.unpack_response(send_value, result_receive.contents[0]) if pack_unpack else result_receive

    def read_server(self, send: Optional[bytearray] = None):
        """使用底层的数据报文来通讯，传入需要发送的消息，返回一条完整的数据指令"""
        result = IOTNetResult()
        with self.lock:
            # 获取有用的网络通道，如果没有，就建立新的连接
            result_socket = self.get_socket()
            if result_socket.is_success is False:
                self.socket_error = True
                result.copy(result_socket)
                return result

            read = self.read_from_socket(result_socket.contents[0], send)
            if read.is_success:
                self.socket_error = False
                result.is_success = read.is_success
                result.contents[0] = read.contents[0]
                result.msg = f"Success"
                self.logging(content=f"recv {len(read.contents[0])}: [{format_bytes(read.contents[0])}]")
            else:
                result.copy(read)

        if self.is_persistent is False:
            self.close_socket(result_socket.contents[0])
        return result

    def receive_client(self, s: socket):
        """在其他指定的套接字上，使用报文来通讯，传入需要发送的消息，返回一条完整的数据指令"""
        initi = self.extra_on_connect(s)
        if initi.is_success is False:
            if s is not None:
                self.close_socket(s)
            return initi

        while self.iot_socket is not None:
            net_msg = self.get_net_msg()

            # 接收数据信息
            result_receive = self.receive_msg(s, net_msg)
            if result_receive.is_success is False:
                return IOTNetResult(code=result_receive.code, msg=result_receive.msg)

            # 拼接结果数据
            self.logging(content=f"recv {len(result_receive.contents[0])}: [{format_bytes(result_receive.contents[0])}]")
            self.extra_on_receive(s, result_receive.contents[0])



import logging
import time
from typing import Callable, NamedTuple, Union

import paho.mqtt.client as mqtt

'''
    pip install paho-mqtt
'''


class MQTTConfig(NamedTuple):
    HOST: str
    PORT: int = 1883    # 使用SSL/TLS的默认端口是 8883
    USER: str = None
    PSW: str = None
    TOPIC: str = ''
    CLIENT_ID: str = ''
    KEEP_ALIVE: int = 60
    SHORT_CONNECT: bool = False  # 短连接模式
    MIN_RECONNECT_DELAY: int = 1
    MAX_RECONNECT_DELAY: int = 120


class MQTTAccessor:

    def __init__(self, config: MQTTConfig):
        self.config = config
        self.client = None
        self.callbacks = {}

    def __del__(self):
        self.close()
    
    def exit(self):
        self.close()

    def connack_string(self, connack_code: int):
        """Return the string associated with a CONNACK result."""
        if connack_code == 0:
            return "success"
        elif connack_code == 1:
            return "unacceptable protocol version."
        elif connack_code == 2:
            return "identifier rejected."
        elif connack_code == 3:
            return "broker unavailable."
        elif connack_code == 4:
            return "bad user name or password."
        elif connack_code == 5:
            return "not authorised."
        else:
            return f"unknown reason({connack_code})."

    def on_connect(self, client, userdata, flags, rc):
        if int(str(rc)) == 0:
            self.set_call_result('debug_log', content=f"connect success({flags})")
        else:   # 连接失败
            self.set_call_result('debug_log', content=f"connect fail({self.connack_string(rc)})({flags})")

    def on_message(self, client, userdata, message):
        self.set_call_result('receive_data', topic=message.topic, payload=message.payload)
        self.set_call_result('debug_log', content=f"recv message({message.topic} {message.payload})")

    def on_log(self, mqttc, userdata, level, string):
        pass

    def on_subscribe(self, client, userdata, mid, granted_qos):
        self.set_call_result('debug_log', content=f"subscribe(mid: {mid}) success")

    def on_disconnect(self, client, userdata, rc):
        self.set_call_result('debug_log', content=f"disconect({self.connack_string(rc)})")

    def on_publish(self, client, userdata, mid):
        self.set_call_result('debug_log', content=f"publish(mid: {mid}) success")

    def close(self):
        try:
            if self.client:
                self.client.disconnect()
                self.client.loop_stop()
                self.set_call_result('debug_log', content=f"close success")
        finally:
            self.client = None

    def get_client(self, client_mode='sub'):
        if self.client is None:
            clean_session = self.config.SHORT_CONNECT  # False表示要建立一个持久性会话
            if self.config.CLIENT_ID is None or len(self.config.CLIENT_ID) == 0:
                client = mqtt.Client(client_id='', clean_session=True)
            else:
                client = mqtt.Client(self.config.CLIENT_ID, clean_session=clean_session)
            if self.config.USER is not None and self.config.PSW is not None and len(self.config.USER) > 0:
                client.username_pw_set(self.config.USER, self.config.PSW)
            client.on_connect = self.on_connect
            client.on_message = self.on_message
            client.on_subscribe = self.on_subscribe
            client.on_disconnect = self.on_disconnect
            client.on_publish = self.on_publish
            client.on_log = self.on_log
            client.connect(self.config.HOST, self.config.PORT, self.config.KEEP_ALIVE)
            client.reconnect_delay_set(self.config.MIN_RECONNECT_DELAY, self.config.MAX_RECONNECT_DELAY)
            if client_mode == 'pub':
                client.loop_start()

                wait = 5
                while not client.is_connected() and wait > 0:    # 等待连接
                    time.sleep(1)
                    wait = wait - 1

                if client.is_connected() is False:
                    client.disconnect()
                    client.loop_stop()
                    raise Exception(f"connect fail")
            self.client = client
        return self.client

    def publish_topic(self, topic: str, message: str, qos: int = 0) -> bool:
        client = self.get_client('pub')
        if client is not None:
            connect_status = False
            if client._state != 2:
                connect_status = True
            else:
                self.set_call_result('debug_log', content=f"client status({client._state}) fail")
            try:
                if connect_status is True:
                    info = client.publish(topic, payload=message, qos=qos)
                    info.wait_for_publish()
                    if info.rc == mqtt.MQTT_ERR_SUCCESS:
                        return True
                    elif info.rc == mqtt.MQTT_ERR_NO_CONN:
                        connect_status = False
                    raise Exception(f"publish fail({info})")
                else:
                    raise Exception(f"invalid state({client._state})")
            finally:
                if self.config.SHORT_CONNECT is True or connect_status is False or self.is_connected() is False:
                    self.close()
        else:
            raise Exception(f"no client")

    def subscribe_topics(self, topics: Union[str, list], qos: int = 0, retry_interval: int = 10, callback: Callable = None):
        self.callbacks['receive_data'] = callback
        while True:
            try:
                client = self.get_client()
                if client:
                    if isinstance(topics, str):
                        client.subscribe(topics, qos=qos)
                    elif isinstance(topics, list):
                        for topic in topics:
                            client.subscribe(topic, qos=qos)
                    client.loop_forever()
            except Exception as e:
                logging.error(f'subscribe topics={topics} fail({e.__str__()})', exc_info=True)
            finally:
                self.close()
            time.sleep(retry_interval)

    def is_connected(self) -> bool:
        if self.client and self.client.is_connected():
            return True
        return False

    def enable_logging(self, enable_log: bool, callback: Callable = None):
        if enable_log is True:
            self.callbacks['debug_log'] = callback
        else:
            self.callbacks['debug_log'] = None

    def set_call_result(self, call_method: str, **kwargs):
        if isinstance(self.callbacks, dict):
            call_method = self.callbacks.get(call_method)
            if isinstance(call_method, Callable):
                call_method(**kwargs)



import os
import time

from robertcommonio.system.io.ftp import FtpConfig, FtpAccessor
from robertcommonbasic.basic.cls.utils import daemon_thread
from typing import Callable

def test_client():
    time.sleep(10)
    access = FtpAccessor(FtpConfig(HOST='127.0.0.1', PORT=5001, USER='admin', PSW='123456'))
    client = access.get_client()
    assert access.download_file(client, 'test.ini', 'testv1.ini') == True

@daemon_thread
def test_server():
    #admin,123456,elramwM,;

    access = FtpAccessor(FtpConfig(HOST='0.0.0.0', PORT=5001, USERS=[{'name': 'admin', 'psw': '123456', 'folder': os.getcwd(), 'role': 'elramwM'}]))
    access.start_server()

@daemon_thread
def test_thread_server(func: str):
    access = FtpAccessor(FtpConfig(HOST='0.0.0.0', PORT=5001, USERS=[{'name': 'admin', 'psw': '123456', 'folder': os.getcwd(), 'role': 'elramwM'}]))
    access.__getattribute__(func)()


test_thread_server('start_server').start()

test_client()



from robertcommonio.system.io.ini import INIConfig, INIAccessor

def test_ini():
    config = INIConfig(PATH='test.ini', PARAMS={'HTTP_SERVER': {'HTTP_HOST': '0.0.0.0', 'HTTP_PORT': 5002}})
    accessor = INIAccessor(config)
    assert accessor.get('HTTP_HOST', str) != '0.0.0.0'
    accessor.set('HTTP_HOST', '127.0.0.1')
    assert accessor.get('HTTP_HOST', str) == '127.0.0.1'
    print()

test_ini()


import json
import time
from robertcommonio.system.io.mqtt import MQTTConfig, MQTTAccessor
from datetime import datetime

HOST = 'mqtt.smartbeop.com'
PORT = 1884
client_id = ''
TOPIC = 'clp' #'SUBSTATION/MASTER/200120-1/S_SNT_DA' #'SUBSTATION/MASTER/200120-1/S_SNT_DAT'
USER = 'clp'
PSW = '9T3QV7hdjcHNIHpt'

all_values = {}

def call_back(topic: str, payload: bytes):
    try:
        values = json.loads(payload.decode())
        print(values)
        if isinstance(values, list):
            for value in values:
                if 'ID' in value.keys() and 'TS' in value.keys() and 'ST' in value.keys() and  'VR' in value.keys():
                    time = datetime.utcfromtimestamp(int(value.get('TS'))).replace(second=0).strftime("%Y-%m-%d %H:%M:%S")
                    if time not in all_values.keys():
                        all_values[time] = {}
                    all_values[time][f"{value.get('ID')}_{value.get('ST')}"] = eval(value.get('VR'))[0]

            print(f"{all_values}")
    except Exception as e:
        print(e.__str__())

def test_pub():
    accessor = MQTTAccessor(MQTTConfig(HOST=HOST, PORT=PORT, USER=USER, PSW=PSW, TOPIC=TOPIC, CLIENT_ID=client_id, KEEP_ALIVE=60))
    while True:
        accessor.publish_topic(TOPIC, datetime.now().strftime('%H:%M:%S'), 0)
        time.sleep(2)

def test_sub():
    accessor = MQTTAccessor(MQTTConfig(HOST=HOST, PORT=PORT, USER=USER, PSW=PSW, TOPIC=TOPIC, CLIENT_ID=client_id, KEEP_ALIVE=60))
    accessor.subscribe_topics(['clp', 'clp/s1'], 0, 10, call_back)

test_sub()



from robertcommonio.system.io.sqlalche import SQLAlCheAccessor
import sqlite3
import cx_Oracle
import os
os.environ['NLS_LANG'] = 'SIMPLIFIED CHINESE_CHINA.UTF8'

def test_sqlite():
    accessor = SQLAlCheAccessor()
    accessor.add_engine('sqlite0', 'sqlite+pysqlite:///:memory:')
    accessor.execute_sql('sqlite0', 'CREATE TABLE some_table (x int, y int)')
    accessor.execute_sql('sqlite0', 'INSERT INTO some_table (x,y) VALUES (:x, :y)', [{"x": 1, "y": 1}, {"x": 2, "y": 4}])
    print(accessor.read_sql('sqlite0', 'SELECT * FROM some_table'))

def test_sqlite1():
    accessor = SQLAlCheAccessor()
    accessor.add_engine('sqlite0', 'sqlite:///config.db')
    accessor.execute_sql('sqlite0', 'CREATE TABLE some_table (x int, y int)')
    accessor.execute_sql('sqlite0', 'INSERT INTO some_table (x,y) VALUES (:x, :y)', [{"x": 1, "y": 1}, {"x": 2, "y": 4}])
    print(accessor.read_sql('sqlite0', 'SELECT * FROM some_table'))

def test_mysql():
    accessor = SQLAlCheAccessor()
    accessor.add_engine('mysql0', 'mysql+pymysql://root:RNB.beop-2013@localhost/beopdata')

    print(accessor.read_sql('mysql0', 'SELECT * FROM some_table'))

    records = [{"A": 1, "B": 1, "C": 5.2, "D": "2021-07-23 00:00:00"}, {"A": 2, "B": 3, "C": 4.2, "D": "2021-07-25 00:00:00"}]
    cmds = accessor.generate_sql_cmds('some_table', records, 'replace', list(records[0].keys()))

    print(accessor.execute_multi_sql('mysql0', cmds))

    print(accessor.read_sql('mysql0', 'SELECT * FROM some_table'))

    record_update = [{"A": 1, "B": 1, "C": 5.2, "D": "2021-07-23 01:00:00"}, {"A": 2, "B": 3, "C": 4.2, "D": "2021-07-25 01:00:00"}]
    cmd_update = accessor.generate_sql_cmds('some_table', record_update, 'update', ["B", "C", "D"], ["A"])

    print(accessor.execute_multi_sql('mysql0', cmd_update))

    print(accessor.read_sql('mysql0', 'SELECT * FROM some_table'))

    print()

def test_oracle():
    accessor = SQLAlCheAccessor()
    accessor.add_engine('oracle0', 'oracle://cy2003:goen_cy2003@10.192.1.216:1521/gc')

    r = accessor.read_sql('oracle0', 'SELECT * FROM action_syn1 where date_out is null')

    records = [{"ID": '1512005962', "CNTR_NO": "123"}, {"ID": '1512005970', "CNTR_NO": "234"}]
    cmds = accessor.generate_sql_cmds('action_syn1', records, 'replace', list(records[0].keys()), ['ID'])

    print(accessor.execute_multi_sql('oracle0', cmds))

    print(accessor.read_sql('mysql0', 'SELECT * FROM some_table'))


    print()

def test_syn_oracle():
    accessor = SQLAlCheAccessor()
    accessor.add_engine('oracle0', 'oracle://XX:XX@10.192.1.250:1521/gc')
    accessor.add_engine('oracle1', 'oracle://XX:XX@10.192.1.216:1521/gc')

    r = accessor.read_sql('oracle0', 'SELECT * FROM action_syn1 where date_out is null')

    records = [{"ID": '1512005962', "CNTR_NO": "123"}, {"ID": '1512005970', "CNTR_NO": "234"}]
    cmds = accessor.generate_sql_cmds('action_syn1', records, 'replace', list(records[0].keys()), ['ID'])

    print(accessor.execute_multi_sql('oracle0', cmds))

    print(accessor.read_sql('mysql0', 'SELECT * FROM some_table'))

test_oracle()


import logging
import json
from typing import Optional, Union, NamedTuple

import requests
from requests.auth import AuthBase
from requests.packages.urllib3.exceptions import InsecureRequestWarning

from robertcommonbasic.basic.error.utils import E_INTERNAL, InputDataError, S_OK
from robertcommonbasic.basic.data.type import JsonType
from robertcommonbasic.basic.validation import input
from robertcommonbasic.basic.log import utils as logutils

# 禁用安全请求警告
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


class Response(NamedTuple):
    success: bool
    code: str
    msg: Optional[str]
    data: Optional[JsonType]
    headers: Optional[dict] = None


class HttpTool:

    def __init__(self):
        pass

    def send_json(self,
                  url: str,
                  data: Union[str, JsonType, None] = None,
                  extra_headers: Optional[dict] = None,
                  auth: Optional[AuthBase] = None,
                  method: Optional[str] = None,
                  encoding: Optional[str] = None,
                  raise_error: Optional[bool] = None,
                  log_input: Optional[bool] = None,
                  log_output: Optional[bool] = None,
                  log_error: Optional[bool] = None,
                  retry: Optional[int] = None,
                  timeout: Optional[int] = None) -> Response:
        return self._send_json(
            url=url, data=data, extra_headers=extra_headers, method=method, auth=auth,
            encoding=encoding, raise_error=raise_error, log_input=log_input,
            log_output=log_output, log_error=log_error, retry=retry,
            timeout=timeout)

    @classmethod
    def _send_json(cls,
                   url: str,
                   data: Union[str, JsonType, None] = None,
                   extra_headers: Optional[dict] = None,
                   method: Optional[str] = None,
                   auth: Optional[AuthBase] = None,
                   encoding: Optional[str] = None,
                   raise_error: Optional[bool] = None,
                   log_input: Optional[bool] = None,
                   log_output: Optional[bool] = None,
                   log_error: Optional[bool] = None,
                   retry: Optional[int] = None,
                   timeout: Optional[int] = None) -> Response:
        args_log = f"encoding={encoding}, raise_error={raise_error}, " \
                   f"log_input={log_input}, log_output={log_output}," \
                   f"log_error={log_error}, retry={retry}, method={method}, " \
                   f"timeout={timeout}"

        max_log_size = 32767
        if extra_headers is None:
            extra_headers = {}
        if not isinstance(extra_headers, dict):
            raise InputDataError(
                f"extra_headers is not a dict: {extra_headers}.")
        raise_error = input.ensure_bool(
            'raise_error', raise_error, default_to=False)
        log_input = input.ensure_bool('log_input', log_input, default_to=False)
        log_output = input.ensure_bool('log_input', log_output, default_to=False)
        log_error = input.ensure_bool('log_error', log_error, default_to=False)
        retry = input.ensure_int('retry', retry, default_to=4)
        method = input.ensure_str('method', method, default_to='POST').upper()
        timeout = input.ensure_int('timeout', timeout, default_to=60)
        encoding = input.ensure_str('encoding', encoding, default_to='utf-8')

        if data is None:
            payload = None
        elif isinstance(data, str):
            payload = data
        else:
            payload = json.dumps(data)

        if log_input:
            logging.info(f'Calling {url} with {args_log} and '
                         f'{payload[:max_log_size] if payload else payload}...')

        last_error = None
        for i in range(0, retry):
            if i > 0 and log_input:
                logging.info(f'Retrying {i}(th) time...', i)
            try:
                headers = {"content-type": f"application/json"}
                headers.update(extra_headers)
                if method == 'GET':
                    response = requests.get(
                        url=url, headers=headers, timeout=timeout, auth=auth, verify=False)
                elif method == 'POST':
                    response = requests.post(
                        url=url, headers=headers, data=payload, timeout=timeout, auth=auth, verify=False)
                else:
                    raise NotImplementedError(
                        f"Method {method} is not supported yet.")

                if response.status_code == 200:
                    response.encoding = encoding
                    if log_output:
                        logging.info(f'Result: {response.text}')
                    return Response(True, S_OK, '', response.text, response.headers)
                raise Exception(f'Unexpected result: {response.status_code} {response.text[:max_log_size]}')
            except Exception as e:
                last_error = e
                if log_error:
                    logutils.log_unhandled_error()

        last_error_msg = f'Failed to post JSON request after all retries! {last_error.__str__()}'
        if log_error:
            logging.error(last_error_msg)
        if raise_error:
            raise last_error
        else:
            return Response(False, E_INTERNAL, last_error.__str__(), None, None)

    def send_request(self, url: str, method: Optional[str] = None, params: Union[str, JsonType, None] = None,
                     data: Union[str, JsonType, None] = None, files: Union[str, JsonType, None] = None,
                     headers: Optional[dict] = None, cookies: Optional[dict] = None,
                     auth: Optional[AuthBase] = None, encoding: Optional[str] = None,
                     raise_error: Optional[bool] = None, log_input: Optional[bool] = None,
                     log_output: Optional[bool] = None, log_error: Optional[bool] = None,
                     retry: Optional[int] = None, timeout: Optional[int] = None) -> Response:
        return self._send_request(
            url=url, method=method, params=params, files=files, data=data, headers=headers, cookies=cookies, auth=auth,
            encoding=encoding, raise_error=raise_error, log_input=log_input,
            log_output=log_output, log_error=log_error, retry=retry,
            timeout=timeout)

    @classmethod
    def _send_request(cls, url: str, method: Optional[str] = None, params: Union[str, JsonType, None] = None,
                      data: Union[str, JsonType, None] = None, files: Union[str, JsonType, None] = None,
                      headers: Optional[dict] = None, cookies: Optional[dict] = None,
                      auth: Optional[AuthBase] = None, encoding: Optional[str] = None,
                      raise_error: Optional[bool] = None, log_input: Optional[bool] = None,
                      log_output: Optional[bool] = None, log_error: Optional[bool] = None,
                      retry: Optional[int] = None, timeout: Optional[int] = None) -> Response:
        args_log = f"encoding={encoding}, raise_error={raise_error}, " \
                   f"log_input={log_input}, log_output={log_output}," \
                   f"log_error={log_error}, retry={retry}, method={method}, " \
                   f"timeout={timeout}"

        max_log_size = 32767
        if headers is None:
            headers = {}
        raise_error = input.ensure_bool('raise_error', raise_error, default_to=False)
        log_input = input.ensure_bool('log_input', log_input, default_to=False)
        log_output = input.ensure_bool('log_input', log_output, default_to=False)
        log_error = input.ensure_bool('log_error', log_error, default_to=False)
        retry = input.ensure_int('retry', retry, default_to=2)
        method = input.ensure_str('method', method, default_to='POST').upper()
        timeout = input.ensure_int('timeout', timeout, default_to=60)
        encoding = input.ensure_str('encoding', encoding, default_to='utf-8')

        payload = data

        if log_input:
            logging.info(f'Calling {url} with {args_log} and {payload[:max_log_size] if payload else payload}...')

        last_error = None
        for i in range(0, retry):
            if i > 0 and log_input:
                logging.info(f'Retrying {i}(th) time...', i)
            try:
                if method == 'GET':
                    response = requests.get(
                        url=url, headers=headers, timeout=timeout, auth=auth, verify=False)
                elif method == 'POST':
                    response = requests.post(
                        url=url, headers=headers, files=files, data=payload, params=params, timeout=timeout, auth=auth, verify=False)
                else:
                    raise NotImplementedError(
                        f"Method {method} is not supported yet.")

                if response.status_code == 200:
                    response.encoding = encoding
                    if log_output:
                        logging.info(f'Result: {response.text}')
                    return Response(True, S_OK, '', response.text, response.headers)
                raise Exception(f'Unexpected result: {response.status_code} {response.text[:max_log_size]}')
            except Exception as e:
                last_error = e
                if log_error:
                    logutils.log_unhandled_error()

        last_error_msg = f'Failed to post JSON request after all retries! {last_error.__str__()}'
        if log_error:
            logging.error(last_error_msg)
        if raise_error:
            raise last_error
        else:
            return Response(False, E_INTERNAL, last_error.__str__(), None, None)



from datetime import datetime, timedelta

from .redis import JsonRedis


class RedisCache:
    __CACHE__ = {}

    @staticmethod
    def load(redis: JsonRedis, func_name: str, name: str, expires: float = 60 * 60 * 2, decode=True):

        subkey = (redis.host, redis.port, func_name)
        data = RedisCache.__load_cache(name, subkey)
        if not data:
            data = RedisCache.__load_redis(redis, func_name, name, decode)
            if data:
                RedisCache.__set_cache(name, subkey, data, expires)

        return data

    @staticmethod
    def load_hm(redis: JsonRedis, name: str, args=(), expires: float = 60 * 60 * 2, decode=True):
        subkey = (redis.host, redis.port, 'hmget') + \
            tuple(args)
        data = RedisCache.__load_cache(name, subkey)
        if not data:
            data = RedisCache.__load_redis(
                redis, 'hmget', name,  decode, list(args))
            if data:
                RedisCache.__set_cache(name, subkey, data, expires)

        return data

    @staticmethod
    def reset(name: str):
        RedisCache.__CACHE__.pop(name, None)

    @staticmethod
    def clear():
        RedisCache.__CACHE__.clear()

    @staticmethod
    def __load_cache(name: str, subkey):
        values = RedisCache.__CACHE__.get(name, None)
        if values is None:
            values = {}
            RedisCache.__CACHE__[name] = values

        v = values.get(subkey)
        if v and v['expires'] > datetime.now():
            return v['data']

        return None

    @staticmethod
    def __set_cache(name: str, subkey, data, expires: float):
        values = RedisCache.__CACHE__.get(name, None)
        if values is None:
            values = {}
            RedisCache.__CACHE__[name] = values

        values[subkey] = {
            'data': data,
            'expires': datetime.now() + timedelta(seconds=expires)
        }

    @staticmethod
    def __load_redis(redis: JsonRedis, func_name: str, name: str, decode: bool, *args):
        m = getattr(redis if decode else redis.conn, func_name, None)
        if not m:
            raise ValueError(f'invalid redis method name: {func_name}')
        return m(name, *args)



import csv
import json
from io import BytesIO
from typing import Any, NamedTuple, Optional, Union
from enum import Enum

import pyzipper
import pandas as pd

from robertcommonbasic.basic.os.file import get_file_folder, rename_file
from robertcommonbasic.basic.os.path import create_dir_if_not_exist


class FileType(Enum):
    AES_ZIP = 'aes_zip'
    CSV = 'csv'
    JSON = 'json'
    Excel = 'excel'


class FileConfig(NamedTuple):
    PATH: Union[str, BytesIO]
    MODE: FileType
    NAME: Optional[str] = None
    PSW: Union[str, list] = None
    FORMAT: str = 'w+'  # r+ 读写  w+ 覆盖 a+ 追加 rb+ 二进制
    ENCODE: str = 'utf-8'
    ENGINE: Optional[str] = None


class FileAccessor:

    def __init__(self, config: FileConfig):
        self.config = config

    def save(self, file_content: Any, file_mode: str = 'w+'):
        create_dir_if_not_exist(get_file_folder(self.config.PATH))

        if self.config.MODE == FileType.AES_ZIP:
            return self.__save_aes_zip(self.config.PATH, self.config.NAME, file_content, self.config.PSW)
        elif self.config.MODE == FileType.CSV:
            return self.__save_csv_file(self.config.PATH, file_content, file_mode, self.config.ENCODE)
        elif self.config.MODE == FileType.JSON:
            return self.__save_json_file(self.config.PATH, file_content, file_mode, self.config.ENCODE)
        elif self.config.MODE == FileType.Excel:
            return self.__save_excel_file(self.config.PATH, file_content, self.config.ENCODE)

    def __save_aes_zip(self, zip_path: str, file_name: str, file_content: Any, zip_pw: Any = None):
        tmp_path = f"{zip_path}.tmp"
        with pyzipper.AESZipFile(tmp_path, 'a', compression=pyzipper.ZIP_DEFLATED) as zip_file:
            if isinstance(zip_pw, str) and len(zip_pw) > 0:
                zip_file.setpassword(zip_pw.encode('utf-8'))
            zip_file.setencryption(pyzipper.WZ_AES, nbits=256)
            if isinstance(file_content, str):
                zip_file.writestr(file_name, data=file_content)
        rename_file(tmp_path, zip_path)

    def __save_csv_file(self, file_path: str, file_content: Any, file_mode: str = 'w+', encoding: str = 'utf-8', newline: str = ''):
        tmp_path = f"{file_path}.tmp"
        with open(file_path, mode=file_mode, encoding=encoding) as f:
            if isinstance(file_content, str):
                f.write(file_content)
            elif isinstance(file_content, bytes):
                f.write(file_content)
            elif isinstance(file_content, list):
                for content in file_content:
                    f.write(content)
        rename_file(tmp_path, file_path)

    def __save_json_file(self, file_path: str, file_content: Any, file_mode: str = 'w+', encoding: str = 'utf-8', newline: str = ''):
        tmp_path = f"{file_path}.tmp"
        with open(file_path, mode=file_mode, encoding=encoding, newline=newline) as f:
            writer = csv.DictWriter(f, fieldnames=file_content[0].keys())
            writer.writeheader()
            for row in file_content:
                writer.writerow(row)
        rename_file(tmp_path, file_path)

    def __save_excel_file(self, file_path: str, file_content: Any, encoding: str = 'utf-8'):
        if isinstance(file_content, dict):
            excel_writer = pd.ExcelWriter(file_path, engine='xlsxwriter')
            for k, v in file_content.items():
                df = pd.DataFrame(v)
                df.to_excel(excel_writer,
                            sheet_name=k,
                            index=False,
                            encoding=encoding)
            excel_writer.save()

    def read(self):
        if self.config.MODE == FileType.AES_ZIP:
            return self.__read_aes_zip(self.config.PATH, self.config.NAME, self.config.PSW)
        elif self.config.MODE == FileType.CSV:
            return self.__read_csv_dict(self.config.PATH)
        elif self.config.MODE == FileType.JSON:
            return self.__read_json_dict(self.config.PATH)
        elif self.config.MODE == FileType.Excel:
            return self.__read_excel_dict(self.config.PATH, self.config.NAME, self.config.ENGINE)

    def __read_aes_zip(self, zip_path: str, file_name: str = '', zip_pw: Any = None) -> dict:
        results = {}
        with pyzipper.AESZipFile(zip_path) as zip_file:
            psws = []
            if isinstance(zip_pw, str) and len(zip_pw) > 0:
                psws.append(zip_pw.encode('utf-8'))
            elif isinstance(zip_pw, list):
                for pw in zip_pw:
                    psws.append(pw.encode('utf-8'))
            else:
                psws = [None]

            if file_name is not None and len(file_name) > 0:
                for psw in psws:
                    try:
                        results[file_name] = zip_file.read(file_name, psw)
                        break
                    except Exception as e:
                        if e.__str__().find('password') >= 0:
                            continue
                        else:
                            raise e
            else:
                for file in zip_file.namelist():
                    for psw in psws:
                        try:
                            results[file] = zip_file.read(file, psw)
                            break
                        except Exception as e:
                            if e.__str__().find('password') >= 0:
                                continue
                            else:
                                raise e
        return results

    def __read_csv_dict(self, file_path: str, newline: str = ''):
        with open(file_path, newline=newline) as file:
            reader = csv.DictReader(file)
            return [row for row in reader]

    def __read_excel_dict(self, file_path: str, sheet_name: Optional[str] = None, engine: str = 'openpyxl') -> dict:
        df = pd.read_excel(file_path, engine=engine, sheet_name=sheet_name)
        records = {}
        if isinstance(df, dict):
            for name, _df in df.items():
                records[name] = [{k: v for k, v in v1.items() if v == v and v is not None} for k1, v1 in _df.T.to_dict().items()]
                del _df
        else:
            records[sheet_name] = [
                {k: v for k, v in v1.items() if v == v and v is not None} for k1, v1 in df.T.to_dict().items()
            ]
            del df
        return records

    def __write_csv_row(self, file_path: str, rows: list, file_mode: str = 'w', newline: str = ''):
        with open(file_path, file_mode, newline=newline) as file:
            writer = csv.writer(file, delimiter=' ', quotechar='|', quoting=csv.QUOTE_MINIMAL)
            for row in rows:
                writer.writerow(row)

    def __write_csv_dict(self, file_path: str, rows: dict, file_mode: str = 'w', newline: str = ''):
        with open(file_path, file_mode, newline=newline) as file:
            writer = csv.DictWriter(file, fieldnames=rows[0].keys())
            writer.writeheader()
            for row in rows:
                writer.writerow(row)

    def __read_json_dict(self, file_path: str):
        with open(file_path) as file:
            return json.loads(file.read())



from typing import Optional, Any, Union, List, Dict

import pandas as pd
from sqlalchemy import create_engine, text
from sqlalchemy.pool.impl import QueuePool
from sqlalchemy.schema import CreateTable
from sqlalchemy.ext.declarative import declarative_base


class SQLAlCheAccessor:

    def __init__(self):
        self.engine: dict = {}

    def add_engine(self, engine_name: str, engine_conn: str, engine_pool_class: Any = QueuePool, engine_pool_recycle: int = 60):
        """
                添加一个数据库连接池
                quote_plus(psw)
                @Example:
                    PostgreSQL
                        # default
                        engine = create_engine('postgresql://scott:tiger@localhost/mydatabase')
                        # psycopg2
                        engine = create_engine('postgresql+psycopg2://scott:tiger@localhost/mydatabase')
                        # pg8000
                        engine = create_engine('postgresql+pg8000://scott:tiger@localhost/mydatabase')

                    MySQL
                        # default
                        engine = create_engine('mysql://scott:tiger@localhost/foo')
                        # mysqlclient (a maintained fork of MySQL-Python)
                        engine = create_engine('mysql+mysqldb://scott:tiger@localhost/foo')
                        # PyMySQL
                        engine = create_engine('mysql+pymysql://scott:tiger@localhost/foo')
                        # mysql+pymysql://root:XXX@localhost:3306/foo?charset=utf8

                    Oracle
                        engine = create_engine('oracle://scott:tiger@127.0.0.1:1521/sidname')
                        engine = create_engine('oracle+cx_oracle://scott:tiger@tnsname')

                    SQL Server
                        # pyodbc
                        engine = create_engine('mssql+pyodbc://scott:tiger@mydsn')
                        # pymssql
                        engine = create_engine('mssql+pymssql://scott:tiger@hostname:port/dbname')

                    SQLite
                        engine = create_engine('sqlite:///foo.db')  # 相对地址
                        engine = create_engine('sqlite:///C:\\path\\to\\foo.db')    # 绝对地址

            """
        if engine_name not in self.engine.keys():
            engine = create_engine(engine_conn, poolclass=engine_pool_class, pool_recycle=engine_pool_recycle)
            if engine:
                self.engine[engine_name] = engine
        return self.engine.get(engine_name)

    def get_engine(self, engine_name: str):
        return self.engine.get(engine_name)

    def to_dict(self, records: Union[List[Optional[Dict]], Dict]):
        if isinstance(records, Dict):
            values = {}
            for k, v in records.items():
                values[k] = v
            return values
        elif isinstance(records, List):
            values = []
            for record in records:
                value = {}
                for k, v in record.items():
                    value[k] = v
                values.append(value)
            return values
        return records

    def read_sql(self, engine_name: str, sql_cmd: str, to_dict: bool = True):
        with self.get_engine(engine_name).connect() as conn:
            records = conn.execute(sql_cmd).mappings().all()
            if to_dict is True:
                return self.to_dict(records)
            return records

    def read_sql_dataframe(self, engine_name: str, sql_cmd: str):
        with self.get_engine(engine_name).connect() as conn:
            return pd.read_sql_query(sql_cmd, conn)

    def execute_sql(self, engine_name: str, sql_cmd: str, params: Optional[list] = None):
        with self.get_engine(engine_name).connect() as conn:
            return conn.execute(text(sql_cmd), params).rowcount

    def execute_multi_sql(self, engine_name: str, cmd_tuple: list):
        with self.get_engine(engine_name).begin() as conn:
            rowcount = 0
            for cmd in cmd_tuple:
                rowcount = conn.execute(text(cmd[0]), cmd[1]).rowcount + rowcount
            return rowcount

    def get_table_struct(self, engine_name: str, table_name: str):
        engine = self.get_engine(engine_name)
        if engine is not None:
            base = declarative_base()
            base.metadata.reflect(engine)
            # 获取原表对象
            old_table = base.metadata.tables[table_name]
            # 获取原表建表语句
            crate_sql = str(CreateTable(old_table))
            base.metadata.clear()
            return crate_sql
        return None

    def generate_sql_format(self, table_name: str, action: str = 'replace', colums: Optional[list] = None, filter_column: Optional[list] = None, add_syntax: bool = False):
        filed_format = '{}' if add_syntax is False else '`{}`'
        filed_format1 = '{} = :{}' if add_syntax is False else '`{}` = :{}'
        if action in ['replace', 'append']:
            return f"insert into {table_name} ({', '.join(filed_format.format(k) for k in colums)}) VALUES ({', '.join(':{}'.format(k) for k in colums)})"
        elif action == 'update':
            return f"update {table_name} set {', '.join(filed_format1.format(k, k) for k in colums)} where {', '.join(filed_format1.format(k, k) for k in filter_column)}"
        elif action == 'delete':
            return f"delete from {table_name} where {', '.join(filed_format1.format(k, k) for k in filter_column)}"

    def generate_sql_cmds(self, table_name: str, records: list, action: str = 'replace', colums: list = None, filter_column: list = None):
        cmds = []
        if len(records) > 0:
            if action == 'replace':
                if filter_column is None:
                    cmds.append((f"delete from {table_name}", None))
                else:
                    cmds.append((self.generate_sql_format(table_name, 'delete', None, filter_column), records))

            cmds.append((self.generate_sql_format(table_name, action, records[0].keys() if colums is None else colums, filter_column), records))
        return cmds

    def copy_table_struct(self, engine_name: str, table_name: str, new_table_name: str):
        crate_sql = self.get_table_struct(engine_name, table_name)
        if crate_sql is not None:
            return self.execute_sql(engine_name, crate_sql.replace("CREATE TABLE " + table_name, "CREATE TABLE if not exists " + new_table_name))

    def syn_table_record(self, engine_name: str, sql_cmd: str, engine_name_new: str, table_name_new: str,  if_exists: str = 'replace'):
        engine = self.get_engine(engine_name)
        engine_new = self.get_engine(engine_name_new)
        if engine and engine_new:
            pd.read_sql(sql_cmd, engine).to_sql(table_name_new, engine_new, if_exists=if_exists)
            return True
        return False



from typing import NamedTuple

import smtplib
import email.mime.multipart
import email.mime.text
from email.utils import formatdate


class EmailConfig(NamedTuple):
    HOST: str
    PORT: int
    USER: str
    PSW: str
    FROM: str


class FileAccessor:

    def __init__(self, config: EmailConfig):
        self.config = config

    def send_email(self, mail_tos: str, mail_content: str, mail_title: str = ''):
        smtp_msg = email.mime.multipart.MIMEMultipart()
        smtp_msg['from'] = self.config.FROM
        smtp_msg['to'] = mail_tos
        smtp_msg['subject'] = mail_title
        smtp_msg['date'] = formatdate(localtime=True)
        smtp_msg.attach(email.mime.text.MIMEText(mail_content))

        smtp_handle = smtplib.SMTP()
        smtp_handle.connect(self.config.HOST, self.config.PORT)
        smtp_handle.starttls()
        smtp_handle.login(self.config.USER, self.config.PSW)
        smtp_handle.sendmail(self.config.FROM, mail_tos.split(';'), smtp_msg.as_string())
        smtp_handle.quit()



import json
import logging
import os
import tempfile
from json import JSONDecodeError
from typing import List, NamedTuple

import oss2
from oss2.iterators import ObjectIteratorV2

from robertcommonbasic.basic.validation import input as input_checker


class OSSConfig(NamedTuple):
    ENDPOINT: str
    ACCESS_KEY_ID: str
    ACCESS_KEY_SECRET: str


class OSSAccessor:
    def __init__(self, config: OSSConfig):
        self.endpoint = input_checker.ensure_not_none_str(
            'ENDPOINT', config.ENDPOINT)
        self.access_key_id = input_checker.ensure_not_none_str(
            'ACCESS_KEY_ID', config.ACCESS_KEY_ID)
        self.access_key_secret = input_checker.ensure_not_none_str(
            'ACCESS_KEY_SECRET', config.ACCESS_KEY_SECRET)

    def _get_bucket(self, bucket_name: str):
        bucket = oss2.Bucket(
            oss2.Auth(self.access_key_id, self.access_key_secret),
            self.endpoint, bucket_name)
        return bucket

    def upload_by_file_path(self,
                            bucket_name: str,
                            oss_path: str,
                            file_path: str,
                            headers: dict = {}):
        # data is string or file stream
        logging.info(
            f'upload oss file to bucket_name={bucket_name}, oss_path={oss_path}'
        )
        bucket = self._get_bucket(bucket_name)
        with open(oss2.to_unicode(file_path), 'rb') as f:
            bucket.put_object(oss_path, f, headers=headers)
        return

    def get_oss_data_file(self, bucket_name, oss_path):
        bucket = self._get_bucket(bucket_name)
        try:
            return bucket.get_object(oss_path)
        except oss2.exceptions.NoSuchKey:
            logging.error(
                f'bucket_name={bucket_name}, oss path={oss_path} not exists')
            return None
        except Exception as e:
            logging.error(
                f'bucket_name={bucket_name}, '
                f'oss path={oss_path} can not fetch, {e}',
                exc_info=True)
            return None

    def get_oss_data_json(self, bucket_name, oss_path):
        try:
            f = self.get_oss_data_file(bucket_name, oss_path)
            if not f:
                return None
            j = json.load(f)
            return j
        except JSONDecodeError:
            logging.error(f'bucket_name={bucket_name}, '
                          f'oss_path={oss_path}, error json format')
            return None

    def delete_oss_objects(self, bucket_name, oss_path_list: List[str]):
        bucket = self._get_bucket(bucket_name)
        bucket.batch_delete_objects(oss_path_list)
        return

    def oss_path_iterator(self, bucket_name, oss_path_prefix):
        bucket = self._get_bucket(bucket_name)
        for object_info in ObjectIteratorV2(bucket, prefix=oss_path_prefix):
            yield object_info

    def upload_text(self,
                    bucket_name: str,
                    oss_path: str,
                    text: str,
                    mode: str = 'w',
                    headers: dict = {}):
        t = tempfile.NamedTemporaryFile(mode=mode, delete=False)
        t.write(text)
        t.close()
        try:
            self.upload_by_file_path(bucket_name, oss_path, t.name, headers)
        finally:
            os.remove(t.name)
        return

    def upload_json(self,
                    bucket_name: str,
                    oss_path: str,
                    j: dict,
                    headers: dict = {}):
        new_headers = {'Content-Type': 'application/json'}
        new_headers.update(headers)
        text = json.dumps(j, ensure_ascii=False)
        self.upload_text(bucket_name, oss_path, text, headers=new_headers)



import time
from robertcommonio.system.io.rabitmq import RabbitMQConfig, RabbitMQAccessor
from datetime import datetime

HOST = '47.103.96.35'
PORT = 5672
USER = 'admin'
PSW = '!Abc@123'
exchange = 'robert'
exchange_type = 'topic'
routing_key = 'robert_queue'
vir_host = '/'
short_connect = True



def call_back(data):
    print(data)


def test_pub():
    accessor = RabbitMQAccessor(RabbitMQConfig(HOST=HOST, PORT=PORT, USER=USER, PSW=PSW, VIR_HOST=vir_host, SHORT_CONNECT=short_connect))
    while True:
        accessor.publish_exchange(exchange, routing_key, datetime.now().strftime('%H:%M:%S'), exchange_type)
        time.sleep(2)


def test_pub1():
    accessor = RabbitMQAccessor(RabbitMQConfig(HOST=HOST, PORT=PORT, USER=USER, PSW=PSW, VIR_HOST=vir_host, SHORT_CONNECT=short_connect))
    while True:
        accessor.publish_exchange('dc.electric.queue', 'electric', datetime.now().strftime('%H:%M:%S'), 'direct')
        time.sleep(2)


def test_sub():
    pass


test_pub1()



from typing import NamedTuple

from ftplib import FTP
from pyftpdlib.authorizers import DummyAuthorizer
from pyftpdlib.handlers import FTPHandler, ThrottledDTPHandler
from pyftpdlib.servers import FTPServer


'''
server
pip installl pyftpdlib

Read permissions:
"e" = change directory (CWD, CDUP commands)
"l" = list files (LIST, NLST, STAT, MLSD, MLST, SIZE commands)
"r" = retrieve file from the server (RETR command)
Write permissions:

"a" = append data to an existing file (APPE command)
"d" = delete file or directory (DELE, RMD commands)
"f" = rename file or directory (RNFR, RNTO commands)
"m" = create directory (MKD command)
"w" = store a file to the server (STOR, STOU commands)
"M" = change file mode / permission (SITE CHMOD command) New in 0.7.0
"T" = change file modification time (SITE MFMT command) New in 1.5.3

注意Linux模式下 FTP需要使用主动模式连接 一般FTP为被动连接

'''

'''
client
tp.set_debuglevel(2)       #打开调试级别2，显示详细信息
ftp.set_pasv(0)           #0主动模式 1 #被动模式
print ftp.getwelcome()      #打印出欢迎信息
ftp.cmd("xxx/xxx")         #更改远程目录
ftp.set_debuglevel(0)       #关闭调试模式
ftp.quit()              #退出ftp
ftp.dir()               #显示目录下文件信息
ftp.mkd(pathname)          #新建远程目录
ftp.pwd()               #返回当前所在位置
ftp.rmd(dirname)          #删除远程目录
ftp.delete(filename)        #删除远程文件
ftp.rename(fromname, toname)   #将fromname修改名称为toname。
'''


class FTPEventHandler(FTPHandler):

    def on_connect(self):
        pass

    def on_disconnect(self):
        # do something when client disconnects
        pass

    def on_login(self, username):
        # do something when user login
        pass

    def on_logout(self, username):
        # do something when user logs out
        pass

    def on_file_sent(self, file):
        # do something when a file has been sent
        pass

    def on_file_received(self, file):
        # do something when a file has been received
        pass

    def on_incomplete_file_sent(self, file):
        # do something when a file is partially sent
        pass

    def on_incomplete_file_received(self, file):
        # remove partially uploaded files
        pass


class FtpConfig(NamedTuple):
    HOST: str
    PORT: int
    USER: str = ''
    PSW: str = ''
    USERS: list = []        # 用户列表
    ANONY_MOUS: str = ''    # 匿名路径
    UPLOAD_SPEED: int = 300*1024        # 上传速度
    DOWN_SPEED: int = 300*1024        # 下载速度
    CONNECT_LIMIT: int = 150        # 最大连接数
    PER_IP_LIMIT: int = 10          # 单个ip最大连接数
    PASSIVE_PORTS: tuple = (2000, 2200)        # 被动端口范围
    ENABLE_LOG: int = 0
    WELCOME: str = 'Welcome to my ftp'    # 欢迎信息
    DEBUG_LEVEL: int = 0
    BUFFER_SIZE: int = 8192


class FtpAccessor:

    def __init__(self, config: FtpConfig):
        self.config = config
        self.server = None

    def __del__(self):
        self.close_server()

    def exit(self):
        self.close_server()

    # 启动客户端
    def get_client(self):
        client = FTP()
        client.set_debuglevel(self.config.DEBUG_LEVEL)
        client.connect(self.config.HOST, self.config.PORT)
        if len(self.config.USER) > 0:
            client.login(self.config.USER, self.config.PSW)
        return client

    # 启动服务器
    def start_server(self):
        authorizer = DummyAuthorizer()
        for user in self.config.USERS:
            authorizer.add_user(user.get('name'), user.get('psw'), user.get('folder'), user.get('role'))
        if len(self.config.ANONY_MOUS) > 0:
            authorizer.add_anonymous(self.config.ANONY_MOUS)

        throttled = ThrottledDTPHandler
        throttled.read_limit = self.config.DOWN_SPEED
        throttled.write_limit = self.config.DOWN_SPEED

        handler = FTPEventHandler
        handler.authorizer = authorizer
        handler.banner = self.config.WELCOME
        handler.passive_ports = range(self.config.PASSIVE_PORTS[0], self.config.PASSIVE_PORTS[1])

        self.server = FTPServer((self.config.HOST, self.config.PORT), handler)
        self.server.max_cons = self.config.CONNECT_LIMIT
        self.server.max_cons_per_ip = self.config.PER_IP_LIMIT

        self.server.serve_forever()

    def close_client(self, client):
        if client is not None:
            client.quit()

    def close_server(self):
        if self.server is not None:
            self.server.close_all()

    # 切换目录
    def cwd_folder(self, client: FTP, remote_file_folder: str):
        if len(remote_file_folder) > 0:
            client.cwd(remote_file_folder)

    # 下载文件
    def download_file(self, client: FTP, remote_file_name: str, local_file_path: str, remote_file_folder: str = ''):
        self.cwd_folder(client, remote_file_folder)
        result = client.retrbinary(f"RETR {remote_file_name}", open(local_file_path, 'wb').write, self.config.BUFFER_SIZE)
        if result.startswith('226'):
            return True
        raise Exception(f"download fail({result})")

    # 上传文件
    def upload_file(self, client: FTP, remote_file_name: str, local_file_path: str, remote_file_folder: str = ''):
        self.cwd_folder(client, remote_file_folder)
        result = client.storbinary(f"STOR {remote_file_name}", open(local_file_path, 'rb'), self.config.BUFFER_SIZE)
        if result.startswith('226'):
            return True
        raise Exception(f"upload fail({result})")

    # 获取用户列表
    def get_users(self):
        return self.config.USERS

    # 更新用户
    def update_users(self, users: list):
        for user in self.config.USERS:
            if self.server.handler.authorizer.has_user(user.get('name')) is True:
                self.server.handler.authorizer.remove_user(user.get('name'))

        self.config = self.config._replace(USERS=users)

        for user in self.config.USERS:
            self.server.handler.authorizer.add_user(user.get('name'), user.get('psw'), user.get('folder'), user.get('role'))



import logging
import time
from typing import Callable, NamedTuple

import pika


class RabbitMQConfig(NamedTuple):
    HOST: str
    PORT: int = 5672
    USER: str = ''
    PSW: str = ''
    VIR_HOST: str = ''
    SHORT_CONNECT: bool = True  # 短连接模式


class RabbitMQAccessor:

    def __init__(self, config: RabbitMQConfig):
        self.config = config
        self.client = None
        self.callbacks = {}

    def __del__(self):
        self.close()

    def close(self):
        try:
            if self.client:
                self.client.close()
        finally:
            self.client = None

    def get_client(self):
        if self.client is None:
            if self.config.USER is not None and self.config.PSW is not None and len(self.config.USER) > 0:
                credentials = pika.PlainCredentials(self.config.USER, self.config.PSW)
                self.client = pika.BlockingConnection(pika.ConnectionParameters(host=self.config.HOST, port=self.config.PORT, virtual_host=self.config.VIR_HOST, credentials=credentials))
            else:
                self.client = pika.BlockingConnection(pika.ConnectionParameters(host=self.config.HOST, port=self.config.PORT, virtual_host=self.config.VIR_HOST))
        return self.client

    def is_connected(self) -> bool:
        if self.client and self.client.is_open():
            return True
        return False

    def set_call_result(self, call_method: str, **kwargs):
        if isinstance(self.callbacks, dict):
            call_method = self.callbacks.get(call_method)
            if isinstance(call_method, Callable):
                call_method(**kwargs)

    def enable_logging(self, enable_log: bool, callback: Callable = None):
        if enable_log is True:
            self.callbacks['debug_log'] = callback
        else:
            self.callbacks['debug_log'] = None

    def publish_exchange(self, exchange: str, routing_key: str, message: str, exchange_type: str = 'topic'):
        client = self.get_client()
        if client is not None:
            try:
                channel = client.channel()
                channel.exchange_declare(exchange=exchange, exchange_type=exchange_type, durable=True)
                channel.basic_publish(exchange=exchange, routing_key=routing_key, body=message, properties=pika.BasicProperties(delivery_mode=2, ))
                channel.close()
                return True
            finally:
                if self.config.SHORT_CONNECT is True or self.is_connected() is False:
                    self.close()

    def publish_queue(self, queue_name: str, message: str, exchange: str = ''):
        client = self.get_client()
        if client is not None:
            try:
                channel = client.channel()
                channel.queue_declare(queue=queue_name, durable=True)
                channel.basic_publish(exchange=exchange, routing_key=queue_name, body=message, properties=pika.BasicProperties(delivery_mode=2, ))
                channel.close()
                return True
            finally:
                if self.config.SHORT_CONNECT is True or self.is_connected() is False:
                    self.close()

    def consume_exchange(self, exchange: str, exchange_type: str = 'topic', routing_key: str = '#', retry_interval: int = 10, callback: Callable = None):
        self.callbacks['receive_data'] = callback

        def real_callback(c, method, properties, body):
            self.set_call_result('receive_data', body=body)
            c.basic_ack(delivery_tag=method.delivery_tag)

        while True:
            try:
                connection = self.get_client()
                channel = connection.channel()
                channel.exchange_declare(exchange=exchange, exchange_type=exchange_type, durable=True)  # 声明一个Exchange
                channel.queue_declare(queue=exchange, exclusive=True)  # 声明一个队里
                channel.queue_bind(exchange=exchange, queue=exchange, routing_key=routing_key)  # 绑定一个队列
                channel.basic_consume(exchange, real_callback)
                channel.start_consuming()
            except Exception:
                logging.error(f'exchange={exchange} exception', exc_info=True)
            finally:
                self.close()
            time.sleep(retry_interval)

    def consume_queue(self, queue_name: str, retry_interval: int = 10, callback: Callable = None):
        self.callbacks['receive_data'] = callback

        def real_callback(c, method, properties, body):
            self.set_call_result('receive_data', body=body)
            c.basic_ack(delivery_tag=method.delivery_tag)

        while True:
            try:
                connection = self.get_client()
                channel = connection.channel()
                channel.queue_declare(queue=queue_name, durable=True)
                channel.basic_qos(prefetch_count=1)
                channel.basic_consume(queue=queue_name, on_message_callback=real_callback)
                channel.start_consuming()
            except Exception:
                logging.error(f'queue_name={queue_name} exception', exc_info=True)
            finally:
                self.close()
            time.sleep(retry_interval)



from robertcommonio.system.io.file import FileType, FileConfig, FileAccessor
from robertcommonbasic.basic.dt.utils import parse_time
from robertcommonio.system.io.http import HttpTool
import base64
import json
import re
import pyzipper
import pandas as pd
from io import BytesIO

def test_csv():
    accessor = FileAccessor(FileConfig(PATH='E:/test.csv', MODE=FileType.CSV))
    accessor.save('ss')


def test_zip_csv():
    accessor = FileAccessor(FileConfig(PATH=r'E:\Beop\Code\Git\datapushserver\file\real\testdtu\20220711\request_20220711081000.zip', PSW='123456', MODE=FileType.AES_ZIP))
    json_contents = accessor.read()
    for file_name, file_content in json_contents.items():
        if len(file_content) > 0:
            if file_name.endswith('.json'):
                body = json.loads(file_content.decode())
                body = body.get('body')
                file_type = body.get('type', '')
                file_psw = body.get('psw', '')
                file_path = body.get('path', '')
                file_name = body.get('name', '')
                file_content = body.get('content', '')
                accessor = FileAccessor(FileConfig(PATH=BytesIO(base64.b64decode(file_content.encode())), PSW='RNB.beop-2019', MODE=FileType.AES_ZIP))
                json_contents = accessor.read()
                for file_name, file_content in json_contents.items():
                    datas = [p.split(',') for p in file_content.decode('gbk').split(';')]
                    points = {}
                    for data in datas:
                        if len(data) >= 2 and len(data[0]) > 0:
                            points[data[0]] = data[1]

                    value = {'dtuName': 'test2020',
                    'dataType': 0,
                    'dataSource': '',
                    'serverCode': 6,
                    'updateTime': '2022-07-11 08:10:00',
                    'dataStruct': [{'time': '2022-07-11 08:10:00', 'type': 0, 'points': points}]}
                    body = json.dumps(value, ensure_ascii=False).encode("utf-8")
                    rt = HttpTool().send_request(url=f"http://beopservice.smartbeop.com/site/v1.0/update_raw_data_v2", method='POST', headers={'content-type': 'application/json'}, data=body, timeout=30)
                    print(rt)
def test_zip_csv1():
    content = b''
    with open(r'E:\DTU\real\atlanta\20210907\his_20210907095301.zip', 'rb') as f:
        content = f.read()

    accessor = FileAccessor(FileConfig(PATH=BytesIO(content), PSW=['aa', '123456', 'RNB.beop-2019', ''],
                                       MODE=FileType.AES_ZIP))
    results = accessor.read()
    for k, v in results.items():
        print(k)
    results = {}
    with pyzipper.AESZipFile(BytesIO(content)) as zip:
        zip.setpassword('RNB.beop-2019'.encode('utf-8'))
        for file in zip.namelist():
            results[file] = zip.read(file)
    print(results)


def test_excel():
    #import pandas as pd
    #df = pd.read_excel(r'E:\DTU\point\hongqiao_api\point202202221.xls', sheet_name=None)

    accessor = FileAccessor(FileConfig(PATH=r'E:\DTU\point\hongqiao_api\point20220224.xls', MODE=FileType.Excel, NAME=None))
    results = accessor.read()
    for k, v in results.items():
        print(k)
    del accessor

    accessor1 = FileAccessor(FileConfig(PATH=r'E:\DTU\point\hongqiao_api\point20220224_new.xls', MODE=FileType.Excel, NAME=None))
    accessor1.save(file_content=results)
    print()

def test_pcc():
    records = pd.read_csv('E:/PCC_AB_Davis_AirHandlers_Analog (3).csv', keep_default_na=False)
    for index, row in records.iterrows():
        row_value = row.to_dict()
        values = {}
        for k, v in row_value.items():
            if v is not None and len(str(v)) > 0:
                if isinstance(v, str) and v.find(',') > 0:
                    v = v.replace(',', '')
                print(v)

test_zip_csv()


import logging
import os
import functools
import json
import uuid
from typing import Tuple, Type, Optional
from enum import Enum
from bson import ObjectId
from datetime import datetime

import pandas as pd
from flask import Response, has_request_context, make_response, request, send_from_directory

from robertcommonbasic.basic.dt.utils import DATETIME_FMT_FULL, get_datetime
from robertcommonbasic.basic.error.utils import S_OK, E_INTERNAL, RobertError
from robertcommonbasic.basic.validation import input
from robertcommonbasic.basic.log import utils as logutils
from robertcommonbasic.basic.os.path import create_dir_if_not_exist, get_sys_path


class Encoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, ObjectId):
            return str(obj)
        elif isinstance(obj, datetime):
            return obj.strftime(DATETIME_FMT_FULL)
        elif str(obj).lower() == 'nan':
            return None
        elif isinstance(obj, pd.Series):
            return obj.to_json(orient='values')
        elif isinstance(obj, pd.DataFrame):
            return obj.to_json(orient='records')
        elif isinstance(obj, Enum):
            return obj.value
        else:
            return obj


class RequestLogLevel:
    LOG_NULL = 0
    LOG_URL = 1
    LOG_URL_INPUT = 2
    LOG_URL_INPUT_AND_COST = 3
    LOG_ALL = 4


def robert_response(is_success, data, code, msg='') -> Response:
    if not code:
        code = S_OK if is_success else E_INTERNAL
    return Response(
        json.dumps({
            'success': is_success,
            'code': code,
            'data': data,
            'msg': msg
        }, cls=Encoder, ensure_ascii=False),
        mimetype='application/json')


def robert_response_error(data=None, code='0', msg='') -> Response:
    if isinstance(data, Exception):
        logging.error(data)
        data = str(data)
    return robert_response(False, data, code, msg)


def robert_response_success(data=None, code='1') -> Response:
    return robert_response(True, data, code)


ErrorTypes = Tuple[Type[Exception], ...]


def _get_error_response(e: Exception, user_errors: ErrorTypes) -> Response:
    is_robert_error = isinstance(e, RobertError)
    is_user_error = user_errors and isinstance(e, user_errors) or is_robert_error

    msg = getattr(e, 'msg', e.__str__())
    data = dict(error_type=str(type(e)),
                detail=getattr(e, 'data', getattr(e, 'detail', None)))

    if is_user_error:
        code = getattr(e, 'code', E_INTERNAL)
        logging.error(e.__str__())
    else:
        code = E_INTERNAL
        data['inner_code'] = getattr(e, 'code', None)
        # noinspection PyUnusedLocal
        url = request.url or 'N/A'
        logutils.log_unhandled_error()
    return robert_response_error(code=code, msg=msg, data=data)


def response_wrapper(func, tolerable_errors: Optional[ErrorTypes] = None):
    tolerable_errors = input.ensure_tuple_of(
        'tolerable_errors', tolerable_errors, (input.ensure_not_none_of, Type))

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Response:
        try:
            if has_request_context() and 'doc' in request.args:
                rv = make_response(func.__doc__)
                rv.headers['Content-Type'] = 'plain/text'
                return rv
            else:
                rv = func(*args, **kwargs)
                return robert_response_success(data=rv)
        except Exception as e:
            return _get_error_response(e, tolerable_errors)

    return wrapper


def response_wrapper_with_log(log_level: RequestLogLevel = RequestLogLevel.LOG_NULL, folder: Optional[str] = None):

    def inner_response_wrapper(func, tolerable_errors: Optional[ErrorTypes] = None):
        tolerable_errors = input.ensure_tuple_of(
            'tolerable_errors', tolerable_errors, (input.ensure_not_none_of, Type))

        def _save_log(out_time, log: dict):
            if len(log) > 0:
                base_folder = folder if folder is not None else get_sys_path()
                log_folder = f"{base_folder}/file/log/request"
                create_dir_if_not_exist(log_folder)
                log_path = f"{log_folder}/request_{out_time.strftime('%Y_%m_%d')}.log"
                with open(log_path, 'a+') as f:
                    f.write(f"{out_time.strftime('%Y-%m-%d %H:%M:%S.%f')} :: {log}\n")

        def _output_log(id, request, response, time):
            request_log = {}
            out_time = get_datetime()
            if log_level >= RequestLogLevel.LOG_URL:
                request_log['id'] = id
                request_log['path'] = request.path
                request_log['method'] = request.method
                request_log['type'] = request.content_type
                request_log['ip'] = str(request.headers.get('X-Real-IP')) if 'X-Real-IP' in request.headers.keys() else request.remote_addr

            if log_level >= RequestLogLevel.LOG_URL_INPUT:
                request_log['input'] = str(request.data, encoding="utf-8")

            if log_level >= RequestLogLevel.LOG_URL_INPUT_AND_COST:
                request_log['cost'] = '{:.3f}'.format((out_time - time).total_seconds())

            if log_level >= RequestLogLevel.LOG_ALL:
                request_log['ouput'] = str(response.data, encoding="utf-8")

            _save_log(out_time, request_log)

        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Response:
            try:
                if has_request_context() and 'doc' in request.args:
                    rv = make_response(func.__doc__)
                    rv.headers['Content-Type'] = 'plain/text'
                    return rv
                else:
                    #
                    request_start = get_datetime()

                    request_id_key = f"X-Request-ID"
                    request_id_value = str(uuid.uuid4())
                    rv = func(*args, **kwargs)
                    response = robert_response_success(data=rv)
                    response.headers.add(request_id_key, request_id_value)

                    _output_log(request_id_value, request, response, request_start)
                    return response
            except Exception as e:
                return _get_error_response(e, tolerable_errors)

        return wrapper

    return inner_response_wrapper


def file_response_wrapper(func, tolerable_errors: Optional[ErrorTypes] = None):
    tolerable_errors = input.ensure_tuple_of(
        'tolerable_errors', tolerable_errors, (input.ensure_not_none_of, Type))

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Response:
        try:
            file_path = func(*args, **kwargs)
            rsp = send_from_directory(path=file_path, directory=os.path.dirname(file_path), filename=os.path.basename(file_path), as_attachment=True, attachment_filename=os.path.basename(file_path))
            rsp.headers['Content-Type'] = "application/octet-stream; charset=UTF-8"
            return rsp
        except Exception as e:
            return _get_error_response(e, tolerable_errors)

    return wrapper



import configparser
from typing import NamedTuple, Any

from robertcommonbasic.basic.os.file import check_file_exist, get_file_folder
from robertcommonbasic.basic.os.path import create_dir_if_not_exist
from robertcommonbasic.basic.validation import input as input_checker

class INIConfig(NamedTuple):
    PATH: str
    PARAMS: dict

class INIAccessor:

    def __init__(self, config: INIConfig):
        self.file_path: str = config.PATH
        self.default_params: dict = config.PARAMS
        self.config_params: dict = {}
        self.config_sections: dict = {}
        self.load()

    def load(self):
        self.ini_file()

        self.read_file()

    def ini_file(self):
        folder = get_file_folder(self.file_path)
        if len(folder) > 0:
            create_dir_if_not_exist(folder)

        config = configparser.ConfigParser()
        config.read(self.file_path)

        for section_key in self.default_params.keys():
            for param_key in self.default_params.get(section_key).keys():
                if config.has_section(section_key) is False:
                    config.add_section(section_key)
                if config.has_option(section_key, param_key.lower()) is False:
                    config.set(section_key, param_key.lower(), input_checker.ensure_str(param_key, self.default_params.get(section_key)))

        config.write(open(self.file_path, 'w'))

    def read_file(self):
        config = configparser.ConfigParser()
        config.read(self.file_path)

        for section in config.sections():
            for option in config.options(section):
                self.config_sections[option] = section
                self.config_params[option] = config.get(section, option)

    def get(self, param_key: str, param_value_type: Any=str):
        param_key = param_key.lower()
        if param_value_type is int:
            return input_checker.ensure_not_none_int(param_key, self.config_params)
        elif param_value_type is bool:
            return input_checker.ensure_not_none_bool(param_key, self.config_params)
        elif param_value_type is float:
            return input_checker.ensure_not_none_float(param_key, self.config_params)
        elif param_value_type is str:
            return input_checker.ensure_not_none_str(param_key, self.config_params)
        elif param_value_type is list:
            return input_checker.ensure_not_none_list(param_key, self.config_params)
        else:
            return input_checker.ensure_not_none(param_key, self.config_params)

    def set(self, param_key: str, param_value: Any, section_key: str=None):
        param_key = param_key.lower()
        self.config_params[param_key] = param_value

        config = configparser.ConfigParser()
        config.read(self.file_path)

        if section_key is not None:
            if config.has_section(section_key) is False:
                config.add_section(section_key)
            config.set(section_key, param_key, param_value)
            config.write(open(self.file_path, 'w'))
        else:
            if param_key in self.config_sections.keys():
                config.set(self.config_sections.get(param_key), param_key, param_value)
                config.write(open(self.file_path, 'w'))


import time

from typing import Union, Optional
from robertcommonio.system.io.socket import SocketType, SocketConfig, SocketAccessor, SocketHandler, IOTNetwork, IOTNetMessage, IOTNetResult, format_bytes


def call_back(data):
    print(data)


def test_tcp_client():
    config = SocketConfig(MODE=SocketType.TCP_CLIENT, HOST='0.0.0.0', PORT=1000, CALL_BACK={})
    accessor = SocketAccessor(config)
    accessor.start()


def test_tcp_server():
    config = SocketConfig(MODE=SocketType.TCP_SERVER, HOST='0.0.0.0', PORT=9500)
    accessor = SocketAccessor(config)
    accessor.start(False)
    while True:
        time.sleep(1)


def test_s7_client():
    config = SocketConfig(MODE=SocketType.TCP_CLIENT, HOST='0.0.0.0', PORT=1000, CALL_BACK={})
    accessor = SocketAccessor(config)
    accessor.start()


class IOTClientMessage(IOTNetMessage):

    def get_head_length(self) -> int:
        '''协议头数据长度，也即是第一次接收的数据长度'''
        return 2

    def get_content_length(self) -> int:
        '''二次接收的数据长度'''
        if self.heads is not None:
            return self.heads[1]
        else:
            return 0

    def check_head(self) -> bool:
        '''回复报文校验'''
        if self.heads is not None:
            if self.heads[0] == 0xA0:
                return True
            else:
                return False
        else:
            return False

    def check_response(self) -> bool:
        '''回复报文校验'''
        if self.heads is not None:
            if self.heads[0] == 0xA0 and self.create_sum((self.heads + self.contents)[0:-1]) == self.contents[-1]:
                return True
            else:
                return False
        else:
            return False

    def create_sum(self, datas: bytes):
        data = sum(datas) & 0xFF
        data = 255 - data if data > 0 else 255 + data
        return data + 1


class IOTServerMessage(IOTNetMessage):

    def get_head_length(self) -> int:
        '''协议头数据长度，也即是第一次接收的数据长度'''
        return 1024


class IOTClient(IOTNetwork):

    '''timeout为None 不超时'''
    def __init__(self, host: str, port: int, timeout: Optional[int] = 5):
        super().__init__(host, port, timeout)

    def get_net_msg(self):
        return IOTClientMessage()

    def extra_on_connect(self, socket):
        '''连接上服务器后需要进行的二次握手操作'''
        # 返回成功的信号
        return IOTNetResult.create_success()

    def extra_on_disconnect(self, socket):
        return IOTNetResult.create_success()

    def ping(self):
        return self.get_socket().is_success

    def read(self):
        read = self.read_server(None)
        return read


class IOTServer(IOTNetwork):
    '''timeout为None 不超时'''

    def __init__(self, host: str, port: int, timeout: Optional[int] = 5, size: int = 500):
        super().__init__(host, port, timeout, size)

    def get_net_msg(self):
        return IOTServerMessage()

    def extra_on_connect(self, socket):
        '''连接上服务器后需要进行的二次握手操作'''
        # 返回成功的信号
        print(f"connect({socket})")
        return IOTNetResult.create_success()

    def extra_on_close(self, socket):
        print(f"close({socket})")
        return IOTNetResult.create_success()

    def extra_on_disconnect(self, socket):
        print(f"disconnect({socket})")
        return IOTNetResult.create_success()

    def extra_on_receive(self, socket, datas: bytes):
        print(f"recv({format_bytes(datas)})")

    def read(self):
        read = self.read_server(None)
        return read


def test_iot_client():
    client = IOTClient('127.0.0.1', 4001, None)
    while True:
        rt = client.read()
        if rt.is_success is True:
            print(format_bytes(rt.contents[0]))
        else:
            print(rt.msg)


def test_iot_server():
    server = IOTServer('0.0.0.0', 4001, None)
    server.start_server()
    while True:
        time.sleep(1)


test_iot_server()



from robertcommonio.basic.cls.utils import SingleInstance

class TestClass(SingleInstance):

    def __init__(self, **kwargs):
        super(TestClass, self).__init__(**kwargs)
        self.cache = {}

    def set(self, key, value):
        self.cache[key] = value

    def get(self, key):
        return self.cache.get(key)


def get_interface(driver_type):
    """Returns an instance of the interface"""
    module_name = "system.io." + driver_type
    module = __import__(module_name)
    ios = module.io
    sub_module = getattr(ios, driver_type)
    klass = getattr(sub_module, "INIAccessor")
    config = getattr(sub_module, "INIConfig")
    accessor = klass(config(PATH='test.ini', PARAMS={'HTTP_SERVER': {'HTTP_HOST': '0.0.0.0', 'HTTP_PORT': 5002}}))
    assert accessor.get('HTTP_HOST', str) != '0.0.0.0'
    accessor.set('HTTP_HOST', '127.0.0.1')
    assert accessor.get('HTTP_HOST', str) == '127.0.0.1'
    print()
    return accessor

def get_interface1(driver_type):
    """Returns an instance of the interface"""
    module_name = "basic.dt.utils"
    module = __import__(module_name, fromlist=('utils'))
    print(dir(module))
    print(module.get_date())

#get_interface1('ini')

TestClass.getInstance().set('aa', 123)
print(TestClass.getInstance().get('aa'))